import numpy as np
import pandas as pd
import os
from AssayingAnomalies.config import Config
from datetime import datetime
from .getFredData import getFredData


def makeKyleObizhaeva():
    """
    Calculate the monthly Volume over Volatility (VoV) measure for stocks.

    This function computes the VoV based on daily volume and return data.
    It also adjusts the volume data using Consumer Price Index (CPI) for inflation.
    The function filters out the stocks that do not meet the criteria of having
    at least 11 days with non-zero returns and 5 days with positive volume in a month.

    Parameters:
    params (object): An object containing various parameters including paths to
                     the CRSP data folder and other required data.

    Returns:
    pd.DataFrame: A DataFrame containing the monthly VoV measures for each stock.
                  The rows represent months, and the columns represent stocks.
                  Stocks that do not meet the criteria within a month are replaced with NaN.

    Note:
    The function requires the following data files in the specified paths:
    - Daily volume ('dvol.csv')
    - Daily returns ('dret.csv')
    - Daily dates ('ddates.csv')
    - Monthly dates ('dates.csv')
    - Daily prices ('dprc.csv')
    Additionally, it pulls the CPI data from the FRED database for inflation adjustment.
    """

    print(f"Started making Kyle Obizhaeva volume over volatility measure of trading costs. Run began at {datetime.now()}")

    params = Config.load_params()

    crsp_path = params.crspFolder + os.sep
    daily_crsp_path = params.daily_crsp_folder + os.sep

    # Load the necessary variables
    # dvol = pd.read_csv(daily_crsp_path + 'dvol.csv', index_col=0).astype(float)
    # dret = pd.read_csv(daily_crsp_path + 'dret.csv', index_col=0).astype(float)
    dvol = pd.read_parquet(daily_crsp_path + 'dvol.parquet').astype(float)
    dret = pd.read_parquet(daily_crsp_path + 'dret.parquet').astype(float)
    dates = pd.read_csv(crsp_path + 'dates.csv', index_col=0).astype(int)
    # dprc = pd.read_csv(daily_crsp_path + 'dprc.csv', index_col=0).astype(float)
    dprc = pd.read_parquet(daily_crsp_path + 'dprc.parquet').astype(float)

    # Store the constants
    a = 8
    b = 2 / 3
    c = 1 / 3

    # Get dollar volume
    dvol = np.multiply(dvol, np.abs(dprc))

    # Convert 'dvol' and 'dret' indices to datetime
    dret.index = pd.to_datetime(dret.index, format='%Y-%m-%d')
    dvol.index = pd.to_datetime(dvol.index, format='%Y-%m-%d')

    # Pull the inflation series from FRED
    # Get the end date
    finalYear = dates.values.flatten()[-1] // 100
    endDate = str(finalYear) + '-12-31'

    # Pull the CPIAUCNS series from FRED
    fredStruct = getFredData('CPIAUCNS', observation_end=endDate, units='lin', frequency='m', aggregation_method='eop')

    # Use FHT's (2018) normailization
    cpi = pd.DataFrame(fredStruct['value'])
    cpi = cpi / cpi.loc['2000-01-01']

    # Match the dates
    match_dates = pd.to_datetime(dates.values.flatten(), format="%Y%m")
    cpi = cpi.reindex(index=match_dates, fill_value=np.nan)

    # Group by month and calculate standard deviation of daily returns
    monthly_std_dev = dret.resample('M').std()

    # Group by month and calculate the average volume but deflate it according to cpi
    monthly_avg = dvol.resample('M').mean() / np.tile(cpi, len(dret.columns))

    # Create the numerator and denominator
    num = a * monthly_std_dev**b
    den = monthly_avg**c

    # Calculate VoV and set the datetime index
    vov = num / den
    vov.index = cpi.index

    # Create a helper function to check for at least 11 days with non-zero returns and 5 days with positive volume
    def check_applicable_days(group):
        # Criteria for valid days
        valid_volume_days = (group['dvol'] > 0).sum() >= 5
        valid_return_days = (group['dret'].abs() > 0).sum() >= 11

        # Check if each stock meets both criteria
        valid_stocks = valid_volume_days & valid_return_days
        return valid_stocks

    # Combine dvol and dret into a single DataFrame for easy processing
    combined_data = pd.concat([dvol, dret], keys=['dvol', 'dret'], axis=1)
    combined_data.index = dvol.index

    # Group by month and apply the function
    monthly_valid_stocks = combined_data.groupby(pd.Grouper(freq='M')).apply(check_applicable_days)

    # Ensure the indices and columns are aligned with vov
    monthly_valid_stocks.index = vov.index

    # Apply Boolean dataframe to the VoV matrix to remove observations that do not meet the criteria
    vov = vov[monthly_valid_stocks]

    print(f"Finished making Kyle Obizhaeva volume over volatility measure of trading costs at {datetime.now()}")

    return vov




import pandas as pd
import numpy as np
import os
from AssayingAnomalies.config import Config
from multiprocessing import Pool
from .rank_with_nan import rank_with_nan


def testing_fill_missing_tcosts(raw_tcosts, rIVOL, rme):
    """
    Fills missing transaction costs ('tcosts') for stocks on each date by assigning
    the 'tcosts' from their closest match based on Euclidean distances calculated
    from market equity ('rme') and idiosyncratic volatility ('rIVOL').

    This function iterates over each date, identifies stocks with valid 'rme' and
    'rIVOL' values, computes Euclidean distances among these stocks, and updates
    'tcosts' for stocks with missing values using the 'tcosts' of their closest
    valid match.

    Parameters:
    - raw_tcosts (pd.DataFrame): A DataFrame containing transaction costs ('tcosts')
      for stocks across different dates. Rows represent dates, and columns represent
      individual stocks.
    - rIVOL (np.ndarray): A NumPy array containing idiosyncratic volatility values
      for stocks, structured similarly to 'raw_tcosts'.
    - rme (np.ndarray): A NumPy array containing market equity ('rme') values for
      stocks, structured similarly to 'raw_tcosts'.

    Returns:
    - pd.DataFrame: The updated DataFrame with missing 'tcosts' filled in for
      eligible stocks based on their closest matches.

    The function prints updates being made for debugging purposes and includes
    exception handling to report any errors encountered during processing.
    """
    # Iterate over each date
    # for date in range(rme.shape[0])[12:24]:
    for date in range(rme.shape[0]):
        try:
            print(f"Month {date}")

            # Step 1: Filter for valid 'rme' and 'rIVOL' values
            valid_mask = ~np.isnan(rme[date]) & ~np.isnan(rIVOL[date])
            valid_indices = np.where(valid_mask)[0]

            # Step 2: Compute Euclidean distances among valid stocks
            rme_valid = rme[date][valid_mask]
            rIVOL_valid = rIVOL[date][valid_mask]
            rIVOL_diff = rIVOL_valid[:, np.newaxis] - rIVOL_valid
            rme_diff = rme_valid[:, np.newaxis] - rme_valid
            euclidean_distance = np.sqrt(rIVOL_diff ** 2 + rme_diff ** 2)
            np.fill_diagonal(euclidean_distance, np.inf)

            # Step 3: Identify indices to update in 'tcosts'
            # These are indices that are both in 'valid_indices' and have 'NaN' in 'tcosts'
            nan_mask = np.isnan(raw_tcosts.iloc[date].to_numpy())
            indices_to_update = valid_indices[nan_mask[valid_indices]]

            # Step 4: Find closest matches for stocks eligible for update
            min_indices = np.argmin(euclidean_distance, axis=1)
            closest_indices = valid_indices[min_indices]

            # Step 5: Update 'tcosts' for eligible stocks with their closest match's 'tcosts'
            for idx in indices_to_update:
                pos_in_valid_indices = np.where(valid_indices == idx)[0][0]
                closest_idx = closest_indices[pos_in_valid_indices]
                old_value = raw_tcosts.iloc[date, idx]
                update_value = raw_tcosts.iloc[date, closest_idx]
                if not np.isnan(update_value):
                    print(f"Updating {idx}:{old_value} with {closest_idx}:{update_value}")
                    raw_tcosts.iloc[date, idx] = update_value

        except Exception as e:
            print(f"Error processing: {e}")

    return raw_tcosts


params = Config.load_params()
crsp_path = params.crspFolder + os.sep

tcosts_raw = pd.read_parquet(crsp_path + 'tcosts_raw.parquet')

# Load variables we need
me = pd.read_csv(crsp_path + 'me.csv', index_col=0).astype(float)
rme = rank_with_nan(me)

# Load IffVOL3 and rIVOL outside the function
IffVOL3 = pd.read_parquet(crsp_path + 'IffVOL3.parquet')
rIVOL = rank_with_nan(IffVOL3)

# Split tcosts, rme, and rIVOL into chunks
tcosts_chunks = np.array_split(tcosts_raw, os.cpu_count())
rme_chunks = np.array_split(rme, os.cpu_count())
rIVOL_chunks = np.array_split(rIVOL, os.cpu_count())
chunks = zip(tcosts_chunks, rme_chunks, rIVOL_chunks)

with Pool(os.cpu_count()) as pool:
    results = pool.starmap(testing_fill_missing_tcosts, chunks)

tcosts = pd.concat(results)


testing = tcosts.copy()
testing = testing.append(results)




import numpy
import pandas as pd
import numpy as np


def fill_months(data, annual_or_quarterly='quarterly'):

    if annual_or_quarterly.lower() == 'annual':
        persist = 11
    else:
        persist = 2

    if isinstance(data, np.ndarray):
        data = pd.DataFrame(data)

    filled_data = data.copy()
    n_dates = len(data)

    # Create a boolean mask for the data
    mask = data.notna()

    # Forward fill the data, limited by the persistence period
    for shift in range(1, persist + 1):
        shifted_mask = mask.shift(shift, fill_value=False)
        filled_data[shifted_mask] = filled_data.ffill(limit=shift)

    return filled_data


# # Sample Data
# data = pd.DataFrame({
#     'Q1': [1, np.nan, np.nan, np.nan, 5, np.nan, np.nan, np.nan, 9, np.nan, np.nan, np.nan],
#     'Q2': [np.nan, 2, np.nan, np.nan, np.nan, 6, np.nan, np.nan, np.nan, 10, np.nan, np.nan],
# })
#
# filled_data = fill_months(data)



import numpy as np
import pandas as pd

#  #TODO:Note I am currently not allowing the user to request variable number of outputs - they will always receive three. I have also not implemented 'varargin' functionality.
def calcTcosts(tcosts, ind, me, weighting='v', ptfRets=None):
    # Validate inputs
    if not (isinstance(tcosts, np.ndarray) or isinstance(tcosts, pd.DataFrame)):
        raise ValueError("Input 'tcosts' must be a NumPy array or a DataFrame.")

    if not (isinstance(ind, np.ndarray) or isinstance(ind, pd.DataFrame)):
        raise ValueError("Input 'ind' must be a NumPy array or a DataFrame.")

    if not (isinstance(me, np.ndarray) or isinstance(me, pd.DataFrame)):
        raise ValueError("Input 'me' must be a NumPy array or a DataFrame.")

    # Convert DataFrames to NumPy arrays
    if isinstance(tcosts, pd.DataFrame):
        tcosts = tcosts.values

    if isinstance(ind, pd.DataFrame):
        ind = ind.values

    if isinstance(me, pd.DataFrame):
        me = me.values

    # Determine weights based on the 'weighting' input
    if isinstance(weighting, str):
        if weighting.lower() == 'v':
            w = me
        elif weighting.lower() == 'e':
            w = me / me
        else:
            raise ValueError(
                "Optional 'weighting' input should be one of 'v', 'V', 'e', 'E', or a user-defined matrix that has the same dimensions as the 'tcosts' matrix.")
    elif isinstance(weighting, np.ndarray) or isinstance(weighting, pd.DataFrame):
        if weighting.shape == me.shape:
            if isinstance(weighting, pd.DataFrame):
                w = weighting.values
            else:
                w = weighting
        else:
            raise ValueError(
                "Optional 'weighting' input should be one of 'v', 'V', 'e', 'E', or a user-defined matrix that has the same dimensions as the 'tcosts' matrix.")
    else:
        raise ValueError(
            "Optional 'weighting' input should be one of 'v', 'V', 'e', 'E', or a user-defined matrix that has the same dimensions as the 'tcosts' matrix.")

    # Store a few constants
    nMonths, nStocks = tcosts.shape
    nPtfs = int(np.nanmax(ind))

    # Initiate the trading costs, turnover, and change in weight matrices
    ptfTC = np.full((nMonths, nPtfs), np.nan)
    ptfTO = np.full((nMonths, nPtfs), np.nan)
    dWs = np.full((nMonths, nStocks, nPtfs), np.nan)

    # Take all the rebalancing months
    indRebMonths = np.where(np.sum(ind > 0, axis=1) > 0)[0]

    # Figure out the rebalancing frequency
    rebFrequencies = np.diff(indRebMonths, prepend=np.nan)
    rebFrequencies = rebFrequencies[~np.isnan(rebFrequencies)]
    rebFreq = np.argmax(np.bincount(rebFrequencies.astype(int)))

    # Store the increase in market capitalization. This is a shortcut to
    # calculating the gross ex-dividend return
    gretxd = me / np.roll(me, rebFreq, axis=0)
    gretxd[:rebFreq] = np.nan

    # Loop over the portfolios
    for i in range(nPtfs):

        # Initialize the weights for this portfolio
        thisPtfW = np.zeros(ind.shape)

        # Assign the weights from the w matrix
        thisPtfW[ind == (i + 1)] = w[ind == (i + 1)]

        # Calculate the sum of the weights & turn into a matrix
        sumThisPtfW = np.nansum(thisPtfW, axis=1)
        sumThisPtfWMat = np.repeat(sumThisPtfW[:, np.newaxis], nStocks, axis=1)

        # Calculate the weights on the way out
        wOut = thisPtfW / sumThisPtfWMat

        # Lag the outweights by the rebalancing frequency
        lagWOut = np.roll(wOut, rebFreq, axis=0)
        lagWOut[:rebFreq] = 0

        # Allow for increasing the base of the strategy to (1+r). This requires
        # the user to have passed on ptfRets with the correct dimensions
        if ptfRets.shape == (nMonths, nPtfs):
            wOut = wOut * (1 + np.repeat(ptfRets[:, i][:, np.newaxis], nStocks, axis=1))

        # Calculate the weights on the way in: last month's weight times the
        # current month return ex-dividends
        wIn = lagWOut * gretxd

        # Store the change in weights
        dWs[:, :, i] = wOut - wIn

        # Store the turnover and trading costs
        ptfTO[:, i] = np.nansum(np.abs(dWs[:, :, i]), axis=1)
        ptfTC[:, i] = np.nansum(np.abs(dWs[:, :, i] * tcosts), axis=1)

        # Calculate the net dW (dW(Long) - dW(Short))
    dW = dWs[:, :, -1] - dWs[:, :, 0]

    return ptfTC, ptfTO, dW



import numpy as np
from scipy.stats import f


def grs_test_p(results):
    # Initialize a few variables
    pvals = np.ones(2)
    dfs = np.zeros((2, 2))
    Fstats = np.ones(2)

    # Store a few variables
    nPtfs = results['pret'].shape[1]
    nMonths = np.sum(np.isfinite(results['pret'][:, 0]))
    nFactors = results['nFactors']

    # Calculate the GRS stat for the full model
    # Get the degrees of freedom
    dfs[0, :] = [nPtfs, nMonths - nPtfs - nFactors]

    indFinite = np.isfinite(np.sum(results['pret'], axis=1)) & np.isfinite(np.sum(results['factors'], axis=1))

    # Calculate the numerator
    numer = results['alpha'][:nPtfs].T @ np.linalg.inv(np.cov(results['resid'][indFinite][:, :nPtfs].T)) @ \
            results['alpha'][:nPtfs]
    numer = numer / 100**2  # correct for percentage return multiplication

    # Calculate the denominator
    f_mat = results['factors'][indFinite] # the factor time series

    # for i in range(nFactors):
    #     f_mat.append(results['factorLoadings'][i][indFinite])
    #
    # f_mat = np.column_stack(f_mat)
    Ef = np.nanmean(f_mat, axis=0).reshape(-1, 1)
    denom = 1 + Ef.T @ np.linalg.inv(np.cov(f_mat, rowvar=False)) @ Ef

    # The statistic
    Fstats[0] = dfs[0, 1] / dfs[0, 0] * numer / denom

    pvals[0] = 1 - f.cdf(Fstats[0], dfs[0, 0], dfs[0, 1])

    # Calculate the GRS stat for the reduced model
    b = results['factorLoadings'].T
    # x = results['factorLoadings'][indFinite]
    x = f_mat.copy()
    a = np.tile(results['alpha'][:nPtfs], (np.sum(indFinite), 1)) / 100 # TODO:F Since alphas are already off by a factor of 100, this makes them off by another factor of 100. I need to correct this in estFactorRegs
    resid = results['resid'][indFinite, :nPtfs] # TODO:F residuals are not the same as those found in MATLAB.
    resxret = x @ b + a + resid

    # Calculate the numerator
    numer = results['xret'][:nPtfs].T @ np.linalg.inv(np.cov(resxret, rowvar=False)) @ results['xret'][:nPtfs]
    numer = numer / 100**2  # correct for percentage return multiplication
    denom = 1
    dfs[1, :] = [nPtfs, nMonths - nPtfs]

    # The statistic
    Fstats[1] = dfs[1, 1] / dfs[1, 0] * numer / denom
    pvals[1] = 1 - f.cdf(Fstats[1], dfs[1, 0], dfs[1, 1])

    return pvals, Fstats, dfs


# # Test the function
# import scipy.io
# import os
# from AssayingAnomalies import Config
# from AssayingAnomalies.Functions.makeBivSortInd import makeBivSortInd
#
# params = Config()
# params.prompt_user()
# params.make_folders()
# path = r"C:\Users\josh\OneDrive - University at Buffalo\Desktop\Spring_2023\AssayingAnomalies-main\AssayingAnomalies-main\Data" + os.sep
# R = scipy.io.loadmat(path + 'R.mat')['R']
# # R = pd.read_csv(path + 'R.csv', index_col=0)
# me = scipy.io.loadmat(path + 'me.mat')['me']
# NYSE = scipy.io.loadmat(path + 'NYSE.mat')['NYSE']
# ind1 = makeBivSortInd(me, 5, R, 5)
# ret = scipy.io.loadmat(path + 'ret.mat')['ret']
# dates = scipy.io.loadmat(path + os.sep + 'CRSP' + os.sep + 'dates.mat')['dates']
# dates = dates.flatten()
#
# # Run a univariate sort to get the underlying portfolios
# res = runUnivSort(params=params, ret=ret, ind=ind1, mcap=me, dates=dates)



def prt_FMB_results(p, bhat, t):
    """
    Print results from Fama-MacBeth regressions.

    :param p: Parsed input object from runFamaMacBeth.
    :param bhat: Vector of coefficient estimates.
    :param t: Vector of t-statistics.
    """
    # Determine the number of variables
    nX = len(bhat) - 1 if p['noConst'] == 0 else len(bhat)

    # Prepare row names
    rnames = []
    if 'labels' in p and p['labels']:
        for label in p['labels']:
            rnames.append(f"{label:25}")
    else:
        for i in range(nX):
            rnames.append(f"var {i+1:20}")
        if p['noConst'] == 0:
            rnames = ['var 0                     '] + rnames

    # Print the results
    print('-' * 70)
    print('           Results from Fama - MacBeth regressions')
    print('-' * 70)
    header_format = "{:25} {:>12} {:>12}"
    print(header_format.format("", "Coeff", "t-stat"))
    for name, coef, t_stat in zip(rnames, bhat, t):
        row_format = "{:25} {:12.3f} {:12.3f}"
        print(row_format.format(name, coef, t_stat))




import pandas as pd
import numpy as np
from datetime import datetime
import os
from .getFFFactors import getFFFactors
from .makeIndustryClassifications import makeIndustryClassifications
from .makeUniverses import makeUniverses
from .makePastPerformance import makePastPerformance
from .makeIndustryReturns import makeIndustryReturns


def makeCRSPDerivedVariables(params):
    """
    This function creates and stores a structure that contains indicators for small/large caps based on two
    classifications: Fama-French (< or > than NYSE 50th percentile) and Russell ((not) in the top 1000 stocks based on
    market cap).
    ------------------------------------------------------------------------------------------
    Required Inputs:

    ------------------------------------------------------------------------------------------
    Output:
        -None

    ------------------------------------------------------------------------------------------
    Examples:
    makeCRSPDerivedVariables(Params)
    ------------------------------------------------------------------------------------------
    Dependencies:
        -Requires makeCRSPMonthlyData() to have been run
    ------------------------------------------------------------------------------------------
    References
    1. Novy-Marx, R. and M. Velikov, 2023, Assaying anomalies, Working paper.

    """
    # Timekeeping
    print(f"\nNow working on making CRSP derived variables. Run started at {datetime.now()}.\n")

    # crspFolder = r'/home/jlaws13/PycharmProjects/AssayingAnomalies_root/Data/CRSP/'
    crspFolder = params.crspFolder + os.sep

    "Load the necessary data from cloud"
    ret_x_dl = pd.read_csv(crspFolder + 'ret_x_dl.csv', index_col=0)
    permno = pd.read_csv(crspFolder + 'permno.csv', index_col=0)
    dates = pd.read_csv(crspFolder + 'dates.csv', index_col=0)
    crsp_msedelist = pd.read_csv(crspFolder + 'crsp_msedelist.csv', index_col=0)

    "get permnos to drop"
    #get the indices of the permnos in crsp_msedelist.permno  that are NOT in permno
    mask = np.isin(crsp_msedelist.permno, permno, invert=True)
    # get the permnos to remove
    toDrop = crsp_msedelist.permno[mask]

    "drop these permnos from msf_delist dataframe"
    crsp_msedelist = crsp_msedelist.set_index('permno')
    crsp_msedelist = crsp_msedelist.drop(toDrop)
    #changing permnos back to a column instead of index
    crsp_msedelist.reset_index(inplace=True)

    "Fill in the delisting returns. Prefer to put the delisting return in the delisting month but if that's not " \
    "possible, either because theres already a return there or its not within the date range, then we put it in the " \
    "following month. If that's not possible then we don't assign it."
    crsp_msedelist['dlstdt'] = pd.to_datetime(crsp_msedelist['dlstdt']).dt.to_period('M')
    crsp_msedelist['dlstdt'] = crsp_msedelist['dlstdt'].dt.strftime('%Y%m').astype(int)
    crsp_msedelist['permno'] = crsp_msedelist['permno'].astype(str)
    ret = ret_x_dl.copy()
    ret.index = ret.index.astype(int)

    # Iterate through each permno in crsp_msedelist
    for perm in ret.columns:
        print(perm)
        # Get the delist date and return for the permno
        delist_info = crsp_msedelist[crsp_msedelist['permno'] == perm]
        delist_date = delist_info['dlstdt'].iloc[0]
        delist_return = delist_info['dlret'].iloc[0]

        # Find the delisting month and the last month with a return observation
        if delist_date in ret.index:
            r_dt = delist_date
            r_last = ret[perm].last_valid_index()

            # Convert r_dt to a datetime object to get the previous month
            r_dt_datetime = pd.to_datetime(r_dt, format='%Y%m')
            prev_month = int((r_dt_datetime - pd.DateOffset(months=1)).strftime('%Y%m'))

            # Choose where to assign the delisting return
            if pd.isna(ret.at[r_dt, perm]) and pd.notna(ret.at[prev_month, perm]):
                r = r_dt
            elif r_last in ret.index and r_last < ret.index[-1]:
                r = r_last
            else:
                r = None

            # Assign the delisting return if a suitable date is found
            if r is not None:
                ret.at[r, perm] = delist_return

    # Get the delisting return for Kodak. Should equal -0.5381
    c = np.where(permno == 11754)[0][0]
    r = np.where(dates == 201201)[0][0]
    kodak_delist_ret = ret.iloc[r, c]
    print(f'Adjusting for delisting complete. Kodak\'s delisting return was {kodak_delist_ret:.4f}')

    "Save the returns dataframe"
    ret.to_csv(crspFolder + 'ret.csv')

    "load additional objects"
    exchcd = pd.read_csv(crspFolder + 'exchcd.csv', index_col=0)
    vol_x_adj = pd.read_csv(crspFolder + 'vol_x_adj.csv', index_col=0)

    "Adjust NASDAQ volume following Gao and Ritter (2010). See their Appendix B for more details"
    vol = vol_x_adj.copy()

    "Requires too much memory to do entire dataframe at once. Will try and go row by row"
    # Divide by 2 prior to Feb 2001
    for i in range(len(exchcd.columns)):
    # for i in range(10):
    #     print(vol.index[i])
        # Since iterating over columns, I only need the row numbers where both conditions are satisfied for a particular column
        rows = np.where((exchcd.iloc[:, i]==3) & (exchcd.index < 200102))[0]
        vol.iloc[rows, i] = vol.iloc[rows, i]/2

    # Divide by 1.8 for most of 2001
    for i in range(len(exchcd.columns)):
    # for i in range(10):
    #     print(vol.index[i])
        # Since iterating over columns, I only need the row numbers where both conditions are satisfied for a particular column
        rows = np.where((exchcd.iloc[:, i]==3) & (exchcd.index >= 200102) & (exchcd.index < 200201))[0]
        vol.iloc[rows, i] = vol.iloc[rows, i]/1.8

    # Divide by 1.6 for 2002 and 2003
    for i in range(len(exchcd.columns)):
    # for i in range(10):
    #     print(vol.index[i])
        # Since iterating over columns, I only need the row numbers where both conditions are satisfied for a particular column
        rows = np.where((exchcd.iloc[:, i]==3) & (exchcd.index >= 200201) & (exchcd.index < 200401))[0]
        vol.iloc[rows, i] = vol.iloc[rows, i]/1.6

    vol.to_csv(crspFolder + 'vol.csv')

    "Make market capitalization"
    prc = pd.read_csv(crspFolder + 'prc.csv', index_col=0)
    shrout = pd.read_csv(crspFolder + 'shrout.csv', index_col=0)
    me = np.abs(prc) * shrout / 1000
    # replace zeros with nan
    me.replace(0, np.nan, inplace=True)
    me.to_csv(crspFolder + 'me.csv')

    "Make dates for plotting"
    # convert dates into decimal year format. E.g. 20230401 = 2023.25
    # "//" performs integer division which == 'floor' in MATLAB
    # and '%' == 'mod', which calculates the remainder.
    pdates = dates // 100 + dates % 100 / 12
    pdates.to_csv(crspFolder + 'pdates.csv')

    "Make the NYSE indicator variable"
    # This code creates a boolean array of the same size as exchcd that has a value of True where exchcd is equal to 1 and
    # False otherwise. The .astype(int) method is then used to convert the boolean values to integers, so True becomes 1 and
    # False becomes 0. This gives an array with the same shape as exchcd, but with 1s where exchcd is equal to 1 and 0s
    # elsewhere
    NYSE = (exchcd == 1).astype(int)
    NYSE.to_csv(crspFolder + 'NYSE.csv')

    "Download, clean up, and save the Fama-French factors from Ken French's website"
    getFFFactors(params=params)

    "Rename the SIC code variable and create Fama-French industry variables"
    makeIndustryClassifications(params=params)

    "Make & save the industry returns based on FF49 classification"
    FF49 = pd.read_csv(crspFolder + 'FF49.csv', index_col=0).astype(float)
    iFF49ret, iFF49reta = makeIndustryReturns(params, FF49)
    pd.DataFrame(iFF49ret).to_csv(crspFolder + 'iFF49ret.csv')
    pd.DataFrame(iFF49reta).to_csv(crspFolder + 'iFF49reta.csv')

    """Next the function makeUniverses is called."""
    # ---within makeUniverses is makeUnivSortInd, runUnivSort, calcPtfRets, calcTcosts, assignToPtf, estFactorRegs,
    makeUniverses(params=params)

    "Make Share Issuance Variables"
    shrout = pd.read_csv(crspFolder + 'shrout.csv', index_col=0)
    cfacshr = pd.read_csv(crspFolder + 'cfacshr.csv', index_col=0)
    ashrout = shrout * cfacshr
    ashrout.to_csv(crspFolder + 'ashrout.csv')
    dashrout = np.log(ashrout / ashrout.shift(12))
    dashrout.to_csv(crspFolder + 'dashrout.csv')

    "Make Momentumm varibles"
    ret = pd.read_csv(crspFolder + 'ret.csv', index_col=0)

    R = pd.DataFrame(makePastPerformance(ret, 12, 1), columns=ret.columns, index=ret.index)
    R62 = pd.DataFrame(makePastPerformance(ret, 6, 1), columns=ret.columns, index=ret.index)
    R127 = pd.DataFrame(makePastPerformance(ret, 12, 6), columns=ret.columns, index=ret.index)
    R3613 = pd.DataFrame(makePastPerformance(ret, 36, 12), columns=ret.columns, index=ret.index)

    R.to_csv(crspFolder + 'R.csv')
    R62.to_csv(crspFolder + 'R62.csv')
    R127.to_csv(crspFolder + 'R127.csv')
    R3613.to_csv(crspFolder + 'R3613.csv')

    "Test that your data matches up with some reference data"
    # TODO:Fix I haven't completed this section yet.

    "Timekeeping"
    print(f"\nCRSP monthly derived variables run ended at {datetime.now()}.\n")

    return

# makeCRSPDerivedVariables()




import numpy as np
import pandas as pd
import os
from AssayingAnomalies.config import Config
from datetime import datetime

def makeAbdiRanaldi():
    print(f"Making Abdi Ranaldi trading costs. Run began at {datetime.now()}")
    params = Config.load_params()
    crsp_path = params.crspFolder + os.sep
    daily_crsp_path = params.daily_crsp_folder + os.sep

    # Load the necessary variables
    # dbidlo = pd.read_csv(daily_crsp_path + 'dbidlo.csv', index_col=0).astype(float)
    # daskhi = pd.read_csv(daily_crsp_path + 'daskhi.csv', index_col=0).astype(float)
    ddates = pd.read_csv(daily_crsp_path + 'ddates.csv', index_col=0).astype(int)
    # dprc = pd.read_csv(daily_crsp_path + 'dprc.csv', index_col=0).astype(float)
    dbidlo = pd.read_parquet(daily_crsp_path + 'dbidlo.parquet').astype(float)
    daskhi = pd.read_parquet(daily_crsp_path + 'daskhi.parquet').astype(float)
    dprc = pd.read_parquet(daily_crsp_path + 'dprc.parquet').astype(float)
    dcfacpr = pd.read_parquet(daily_crsp_path + 'dcfacpr.parquet').astype(float)
    ret = pd.read_csv(crsp_path + 'ret.csv', index_col=0).astype(float)
    permno = pd.read_csv(crsp_path + 'permno.csv', index_col=0).astype(float)

    # Check for missing data
    # Check for missing columns (permnos) and add them with NaN values
    for df in [dcfacpr, dbidlo, daskhi, dprc, dcfacpr]:
        missing_columns = set(permno.values.flatten()) - set(df.columns.astype(float))

        # Iterate over the list of missing columns
        for missing_column_label in missing_columns:
            # Find the correct position to insert the new column
            column_labels = df.columns.astype(float).tolist()
            insert_position = 0
            for i, label in enumerate(column_labels):
                if label > missing_column_label:
                    insert_position = i
                    break

            # Insert the new column with NaN values at the determined position
            new_column = pd.Series(np.nan, index=df.index)
            df.insert(insert_position, missing_column_label, new_column)

    # Create a copy of raw daily price matrix
    dprc_raw = dprc.copy()
    dhigh = dbidlo.copy()
    dlow = daskhi.copy()

    # Set the daily high and low for days when a stock does not trade to np.nan
    dhigh.iloc[dprc < 0 | np.isnan(dprc)] = np.nan
    dlow.iloc[dprc < 0 | np.isnan(dprc)] = np.nan

    # Mask where the stock didn't trade (dprc < 0 or NaN)
    mask = (dprc_raw < 0) | dprc_raw.isna()

    # Carry over the previous days daily high, low, and close on days when a stock doesn't trade. To achieve this, I
    # forward fill the masked data for dprc, dhigh, and dlow
    dprc.mask(mask).fillna(method='ffill', inplace=True)
    dhigh.mask(mask).fillna(method='ffill', inplace=True)
    dlow.mask(mask).fillna(method='ffill', inplace=True)

    # Take the absolute value of the daily price
    dprc = dprc.abs()

    # Store the midpoints of the low and high for t and tp1 (= t plus one)
    midpoint = (np.log(dlow) + np.log(dhigh)) / 2
    midpoint_tp1 = midpoint.shift(-1)

    # Set the days where the stock does not trade to nan
    dbidlo[dprc_raw < 0 | np.isnan(dprc_raw)] = np.nan
    daskhi[dprc_raw < 0 | np.isnan(dprc_raw)] = np.nan

    # Initiate the close-high-low effective spread measure
    chl = pd.DataFrame(np.nan, index=ddates.values.flatten(), columns=ret.columns)
    chl.index = pd.to_datetime(chl.index, format='%Y%m%d')

    # Calculate the spread
    c_t = np.log(dprc)
    s_hat_t = np.sqrt(np.maximum(4 * (c_t - midpoint) * (c_t - midpoint_tp1), 0))
    s_hat_t[s_hat_t < 0] = 0  # Set negative spreads to 0

    # Set s_hat_t index to datetime
    s_hat_t.index = ddates.values.flatten()
    s_hat_t.index = pd.to_datetime(s_hat_t.index, format='%Y%m%d')

    # Resample and calculate the mean spread for each month
    chl = s_hat_t.resample('M').mean()

    # Define a function to apply to each monthly group
    def check_applicable_days(group):
        criteria = (group['dprc_raw'] > 0) & \
                   group['dprc_raw'].notna() & \
                   group['dbidlo'].notna() & \
                   group['daskhi'].notna() & \
                   (group['daskhi'] - group['dbidlo'] != 0)

        # Count the number of days that meet the criteria for each stock
        valid_days_count = criteria.sum()

        # Check if each stock has at least 12 applicable days
        valid_stocks = valid_days_count >= 12
        return valid_stocks

    # Combine data into a single DataFrame for easy processing
    combined_data = pd.concat([dprc_raw, dbidlo, daskhi], keys=['dprc_raw', 'dbidlo', 'daskhi'], axis=1)
    combined_data.index = s_hat_t.index

    # Group by month and apply the function
    monthly_valid_stocks = combined_data.groupby(pd.Grouper(freq='M')).apply(check_applicable_days)

    chl = chl[monthly_valid_stocks]

    print(f"Finished making Abdi Ranaldi trading costs at {datetime.now()}")

    return chl

# chl = makeAbdiRanaldi(params)



import numpy as np
import pandas as pd
import os
from .makeCOMPUSTATVariables import makeCOMPUSTATVariables
from datetime import datetime

# TODO:Note It says there were 0 cases of permno-years in which companies moved their fiscal year end,
#  which indicates something is off

def mergeCRSPCOMP(params):
    # Timekeeping
    print(f"\nNow working on merging CRSP and COMPUSTAT variables. Run started at {datetime.now()}.\n")

    crspFolder = params.crspFolder + os.sep
    compFolder = params.compFolder + os.sep

    # set year of beginning and end of sample
    SAMPLE_START = params.sample_start
    SAMPLE_END = params.sample_end

    # Read the linking file
    crsp_ccmxpf_lnkhist = pd.read_csv(crspFolder + 'crsp_ccmxpf_lnkhist.csv', index_col=0)

    # Filter based on the CCM link; Leave only link types LC or LU :TODO What does 'LC' and 'LU' represent?
    idxToKeep = crsp_ccmxpf_lnkhist['linktype'].isin(['LC', 'LU'])
    crsp_ccmxpf_lnkhist = crsp_ccmxpf_lnkhist[idxToKeep]

    # Remove unwanted variables (columns)
    crsp_ccmxpf_lnkhist = crsp_ccmxpf_lnkhist[['lpermno', 'gvkey', 'linkdt', 'linkenddt']]

    # Replace the missing linkeddt with the chosen end of sample
    indNatEndDate = pd.isna(crsp_ccmxpf_lnkhist['linkenddt'])
    crsp_ccmxpf_lnkhist.loc[indNatEndDate, 'linkenddt'] = datetime(SAMPLE_END, 12, 31)

    # Load the annual COMPUSTAT file
    # comp_funda = pd.read_csv(compFolder + 'comp_funda.csv',  index_col=0, nrows=1000)
    comp_funda = pd.read_csv(compFolder + 'comp_funda.csv', index_col=0)
    comp_funda['datadate'] = pd.to_datetime(comp_funda['datadate'], format='%Y-%m-%d')  # Convert to datetime format

    # First create a boolean series indicating which rows don't satisfy our criteria
    idxToDrop = (comp_funda['datadate'] > datetime(SAMPLE_END, 12, 31)) | \
                (comp_funda['datadate'] < datetime(SAMPLE_START, 1, 1))
    # Then drop the rows
    comp_funda.drop(index=comp_funda[idxToDrop].index, inplace=True)

    # Merge the dataframes, making sure to specify we want to keep all values from comp_funda and crsp_ccmxpf_lnkhist. Setting sort=False increases performance
    comp_funda_linked = pd.merge(comp_funda, crsp_ccmxpf_lnkhist, how='outer', on='gvkey', sort=False)

    # Fiscal period end date must be within link date range. Need to convert the string to a datetime object before doing the comparison.
    comp_funda_linked['linkenddt'] = pd.to_datetime(comp_funda_linked['linkenddt'], format='%Y-%m-%d')
    comp_funda_linked['linkdt'] = pd.to_datetime(comp_funda_linked['linkdt'], format='%Y-%m-%d')
    idxToDrop = (comp_funda_linked['datadate'] > comp_funda_linked['linkenddt']) | \
                (comp_funda_linked['datadate'] < comp_funda_linked['linkdt'])
    comp_funda_linked = comp_funda_linked.loc[~idxToDrop]

    # Must have permno
    indxToDrop = comp_funda_linked.lpermno.isna()
    comp_funda_linked = comp_funda_linked.loc[~indxToDrop]

    # Create a new 'dates' column in YYYYMM format. The '.dt' lets us access the 'year' attribute of the 'datadate' column
    comp_funda_linked['dates'] = (comp_funda_linked['datadate'].dt.year + 1) * 100 + 6

    # Drop a few variables and change the name of permno column
    comp_funda_linked.drop(columns=['gvkey', 'linkdt', 'linkenddt'], inplace=True)
    comp_funda_linked.rename(columns={'lpermno': 'permno'}, inplace=True)

    # Note: there are cases where a company changes its fiscal year end, which
    # results in more than one observation per permno-year. See, e.g., year
    # 1969 for permno 10006. We'll deal with those here by keeping only the
    # data from the fiscal year that happens later in the year
    # Sort the table first
    comp_funda_linked.sort_values(by=['permno', 'datadate'], inplace=True)
    nduplicates = (comp_funda_linked.groupby(['permno', 'dates']).size() > 1).sum()
    print(f"There were {nduplicates} cases of permno-years in which companies moved their fiscal year end.")

    # Apply the anonymous function (i.e., leaving the last element) in the first argument to every variable, by grouping them by permno & date
    # Using 'groupby' method to group 'comp_funda_linked' by permno and dates. Then using nested apply methods. The inner one
    # selects the last element of each variable for each row. The outer one selects the last element of each variable for
    # each group.
    # adj_comp_funda_linked = comp_funda_linked.groupby(['permno', 'dates']).apply(lambda x: x.apply(lambda y: y.iloc[-1]))
    # The above line is VERY slow. Below is a faster method
    adj_comp_funda_linked = comp_funda_linked.groupby(['permno', 'dates']).tail(1).copy()

    # The below code doesn't need to be run. The columns titled 'Fun_....' is a Matlab artifact.
    # Fix the names
    # varNames = adj_comp_funda_linked.Properties.VariableNames';
    # clndVarNames = regexprep(varNames,'Fun_','');
    # adj_comp_funda_linked.Properties.VariableNames = clndVarNames;

    # Create the fiscal year end variable in YYYYMMDD format and drop a couple of variables
    adj_comp_funda_linked['FYE'] = (adj_comp_funda_linked['datadate'].dt.year * 10000 +
                                    adj_comp_funda_linked['datadate'].dt.month * 100 +
                                    adj_comp_funda_linked['datadate'].dt.day)
    adj_comp_funda_linked = adj_comp_funda_linked.drop(columns='datadate', axis=1)

    adj_comp_funda_linked.to_csv(compFolder + 'adj_comp_funda_linked.csv')

    # TODO:F This line is redundant but if I remove it then makeCOMPUSTATVariables acts funny.
    # Note! when reading in the file, the column labeled 'unnamed' looks like permno but it is incorrect. The correct
    # permnos are in the permno column
    adj_comp_funda_linked = pd.read_csv(compFolder + 'adj_comp_funda_linked.csv')

    """At this point, the function should call another function named makeCOMPUSTATVariables."""
    makeCOMPUSTATVariables(params=params, data=adj_comp_funda_linked, quarterly_indicator=False)

    "After running makeCOMPUSTATVariables, we need to perform a similar set of operations to the quarterly file"
    # Load the quarterly COMPUSTAT file
    # comp_funda = pd.read_csv(compFolder + 'comp_funda.csv',  index_col=0, nrows=1000)
    comp_fundq = pd.read_csv(compFolder + 'comp_fundq.csv', index_col=0)
    comp_fundq['datadate'] = pd.to_datetime(comp_fundq['datadate'], format='%Y-%m-%d')  # Convert to datetime format

    # First create a boolean series indicating which rows don't satisfy our criteria
    idxToDrop = (comp_fundq.rdq.isna()) | \
                (comp_fundq['datadate'] > datetime(SAMPLE_END, 12, 31)) | \
                (comp_fundq['datadate'] < datetime(SAMPLE_START, 1, 1))
    # Then drop the rows
    comp_fundq.drop(index=comp_fundq[idxToDrop].index, inplace=True)

    # Merge the dataframes, making sure to specify we want to keep all values from comp_funda and crsp_ccmxpf_lnkhist. Setting sort=False increases performance
    comp_fundq_linked = pd.merge(comp_fundq, crsp_ccmxpf_lnkhist, how='outer', on='gvkey', sort=False)

    # Fiscal period end date must be within link date range. Need to convert the string to a datetime object before doing the comparison.
    comp_fundq_linked['linkenddt'] = pd.to_datetime(comp_fundq_linked['linkenddt'], format='%Y-%m-%d')
    comp_fundq_linked['linkdt'] = pd.to_datetime(comp_fundq_linked['linkdt'], format='%Y-%m-%d')
    idxToDrop = (comp_fundq_linked['datadate'] > comp_fundq_linked['linkenddt']) | \
                (comp_fundq_linked['datadate'] < comp_fundq_linked['linkdt'])
    comp_fundq_linked = comp_fundq_linked.loc[~idxToDrop]

    # Must have permno
    indxToDrop = comp_fundq_linked.lpermno.isna()
    comp_fundq_linked = comp_fundq_linked.loc[~indxToDrop]

    # Drop a few variables
    comp_fundq_linked.drop(columns=['gvkey', 'linkdt', 'linkenddt'], inplace=True)

    # Create the dates variable in yyyymm format - assume available at the end of the RDQ month
    # first converts the 'rdq' column to datetime using pd.to_datetime(), and then applies the .dt.strftime() method
    # to format the date as a string in the 'yyyymm' format.
    # The '.dt' lets us access the 'year' attribute of the 'datadate' column
    comp_fundq_linked['dates'] = pd.to_datetime(comp_fundq_linked['rdq']).dt.strftime('%Y%m')

    # Change the name of permno column
    comp_fundq_linked.rename(columns={'lpermno': 'permno'}, inplace=True)

    # Note: there are cases where multiple stock-quarter data points are
    # associated with a single earnings announcement date (RDQ). These could be
    # due to restatements, or delays in announcements. See, e.g., permno 63079
    # announcements for 200708 in comp_fundq_linked:
    # temp=comp_fundq_linked(comp_fundq_linked.permno==63079 & comp_fundq_linked.dates==200708,:);
    # We'll deal with these by leaving the latest available fiscal quarter associated with each RDQ
    comp_fundq_linked.sort_values(by=['permno', 'datadate', 'rdq'], inplace=True)

    # Count duplicates before selecting the last row
    nduplicates = (comp_fundq_linked.groupby(['permno', 'dates']).size() > 1).sum()

    # Apply the anonymous function (i.e., leaving the last element) in the first argument to every variable,
    # by grouping them by permno & date Using 'groupby' method to group 'comp_fundq_linked' by permno and dates. Then
    # using nested apply methods. The inner one selects the last element of each variable for each row. The outer one
    # selects the last element of each variable for each group.
    # adj_comp_fundq_linked = comp_fundq_linked.groupby(['permno', 'dates']).apply(lambda x: x.apply(lambda y: y.iloc[-1]))

    # The above line is VERY slow. Below is a  faster method
    adj_comp_fundq_linked = comp_fundq_linked.groupby(['permno', 'dates']).tail(1).copy()

    print(f"There were {nduplicates} cases of permno-RDQ months associated with multiple quarters.")

    # Create the RDQ and fiscal-quarter-end variables
    adj_comp_fundq_linked['rdq'] = pd.to_datetime(comp_fundq_linked['rdq'])
    adj_comp_fundq_linked['RDQ'] = adj_comp_fundq_linked['rdq'].dt.year * 10000 + \
                                   adj_comp_fundq_linked['rdq'].dt.month * 100 + \
                                   adj_comp_fundq_linked['rdq'].dt.day

    adj_comp_fundq_linked['FQTR'] = adj_comp_fundq_linked['datadate'].dt.year * 10000 + \
                                    adj_comp_fundq_linked['datadate'].dt.quarter * 100 + 99

    adj_comp_fundq_linked = adj_comp_fundq_linked.drop(['rdq', 'datadate'], axis=1)
    adj_comp_fundq_linked.to_csv(crspFolder + 'adj_comp_fundq_linked.csv')

    # TODO:F This line is redundant but if I remove it then makeCOMPUSTATVariables acts funny.
    # Note! when reading in the file, the column labeled 'unnamed' looks like permno but it is incorrect. The correct
    # permnos are in the permno column
    adj_comp_fundq_linked = pd.read_csv(crspFolder + 'adj_comp_fundq_linked.csv')

    """Next I need to call makeCOMPUSTATVariables again to make the quarterly variables"""
    makeCOMPUSTATVariables(params, data=adj_comp_fundq_linked, quarterly_indicator=True)

    # Timekeeping
    print(f"\nFinished merging CRSP and COMPUSTAT variables. Run ended at {datetime.now()}.\n")

    return



import numpy as np
import pandas as pd
from datetime import datetime
import os
from concurrent.futures import ThreadPoolExecutor


def makeCRSPMonthlyData(params):
    # Timekeeping
    print(f"\nNow working on making CRSP monthly data. Run started at {datetime.now()}.")

    "Set path"
    crsp_folder = params.crspFolder + os.sep

    """Load and clean msf dataframe"""
    crsp_msf = pd.read_csv(crsp_folder + 'crsp_msf.csv', index_col=0)

    "Convert dates to YYYYMM format. These are originally month end dates"
    # first convert dates to datatype - date-time"
    crsp_msf['dates'] = pd.to_datetime(crsp_msf.date)
    # then change to YYYYMM format
    crsp_msf['dates'] = crsp_msf['dates'].dt.strftime('%Y%m')
    crsp_msf['dates'] = crsp_msf['dates'].astype(int)

    "Load CRSP montly stock file with share code information"
    crsp_mseexchdates = pd.read_csv(crsp_folder + 'crsp_mseexchdates.csv', index_col=0)
    # only want to keep certain columns
    crsp_mseexchdates = crsp_mseexchdates.loc[:, ['permno', 'namedt', 'nameendt', 'shrcd', 'exchcd', 'siccd']]
    crsp_mseexchdates.duplicated(subset=['permno', 'siccd']).sum()

    # Merge the share code from the header file to crsp_msf
    crsp_msf = crsp_msf.merge(crsp_mseexchdates, how='outer', on='permno')
    duplicates = crsp_msf.duplicated(subset=['dates', 'permno', 'shrcd'])
    duplicates.sum()
    print(f"There are {duplicates.sum()} duplicate shrcd/permno/date values.")

    # creating index to drop
    idxToDrop1 = np.where(crsp_msf.date < crsp_msf.namedt)[0]
    idxToDrop2 = np.where(crsp_msf.date > crsp_msf.nameendt)[0]
    idxToDrop = np.concatenate((idxToDrop1, idxToDrop2))
    idxToDrop = np.sort(idxToDrop)

    # keeping those NOT in idxToDrop
    crsp_msf = crsp_msf.loc[~crsp_msf.index.isin(idxToDrop)]

    # delete dataframe to free up RAM (not sure if this actually does anything)
    del crsp_mseexchdates

    "Check to see if we should only keep share codes 10 or 11 (domestic common equity)"
    if params.domComEqFlag:
        crsp_msf = crsp_msf[crsp_msf['shrcd'].isin([10, 11])].copy()

    "Check to keep only the sample specified in params."
    first_date = params.sample_start*100 + 1
    last_date = (params.sample_end + 1)*100 + 1
    # idx_to_keep = np.where((crsp_msf['dates'] >= first_date) & (crsp_msf['dates'] <= last_date))
    # idx_to_keep = np.sort(idx_to_keep)
    # crsp_msf = crsp_msf.loc[crsp_msf.index.isin(idx_to_keep)]
    crsp_msf = crsp_msf[(crsp_msf['dates'] >= first_date) & (crsp_msf['dates'] <= last_date)].copy()

    "Rename returns to indicate they are without delisting adjustments"
    crsp_msf.rename(columns={'ret': 'ret_x_dl'}, inplace=True)
    "Rename volume to indicate it is without adjustment for NASDAQ"
    crsp_msf.rename(columns={'vol': 'vol_x_adj'}, inplace=True)

    "Save the link file for the COMPUSTAT matrices creation"
    crsp_link = crsp_msf.loc[:, ['permno', 'date']]
    crsp_link.to_csv(crsp_folder + 'crsp_link.csv')
    del crsp_link

    "Create and store the permno and dates vectors"
    permno = np.sort(crsp_msf['permno'].unique())
    dates = np.sort(crsp_msf['dates'].unique())

    "Save permno, dates"
    pd.DataFrame(permno).to_csv(crsp_folder + 'permno.csv')
    pd.DataFrame(dates).to_csv(crsp_folder + 'dates.csv')

    # Choose variables from crsp_msf to convert into matrices
    varNames_crsp_msf = ['shrcd', 'exchcd', 'siccd', 'prc', 'bid', 'ask', 'bidlo', 'askhi', 'vol_x_adj',
                         'shrout', 'cfacpr', 'cfacshr', 'spread', 'retx', 'ret_x_dl']

    if not params.remote_or_not:
        # Create DataFrames for other variables and align them with the reference DataFrame
        for i, var in enumerate(varNames_crsp_msf):
            # Create DataFrame for current variable
            temp_df = pd.pivot_table(crsp_msf, index='dates', columns='permno', values=var)

            # Align columns and rows
            temp_df = temp_df.reindex(index=dates, columns=permno, fill_value=np.nan)

            # # Align columns
            # for col in permno:
            #     if col not in temp_df.columns:
            #         temp_df[col] = np.nan
            #
            # # Align rows
            # for row in dates:
            #     if row not in temp_df.index:
            #         temp_df.loc[row] = np.nan
            #
            # # Sort index and columns to match reference DataFrame
            # temp_df = temp_df.sort_index(axis=0).sort_index(axis=1)
            print(f"{var}: {i+1}/{len(varNames_crsp_msf)} shape is {temp_df.shape}")
            # Save the aligned DataFrame
            temp_df.to_csv(crsp_folder + var + '.csv')
    else:
        def process_and_save_var(permno, dates, crsp_folder, var):
            """
            Process a single variable and save the result to a CSV file.
            """
            print(f"Processing {var}")
            temp_df = pd.pivot_table(crsp_msf, index='dates', columns='permno', values=var)

            # Align columns and rows
            temp_df = temp_df.reindex(index=dates, columns=permno, fill_value=np.nan)

            # # Align columns
            # for col in permno_list:
            #     if col not in temp_df.columns:
            #         temp_df[col] = np.nan
            #
            # # Align rows
            # for row in dates_list:
            #     if row not in temp_df.index:
            #         temp_df.loc[row] = np.nan
            #
            # # Sort index and columns to match reference DataFrame
            # temp_df = temp_df.sort_index(axis=0).sort_index(axis=1)

            # Save the aligned DataFrame
            temp_df.to_csv(crsp_folder + var + '.csv')
            print(f"Finished processing {var} with shape {temp_df.shape}")

        with ThreadPoolExecutor(max_workers=params.num_cpus) as executor:
            # Submit tasks to the thread pool
            futures = [executor.submit(process_and_save_var, permno, dates, crsp_folder, var) for var in varNames_crsp_msf]

            # wait for each task to complete and print any errors if they exist.
            for future in futures:
                future.result()  # Waits for the individual task to complete

    # Timekeeping
    print(f"\nFinished making CRSP monthly data. Run ended at {datetime.now()}.\n")

    return


# makeCRSPMonthlyData(params)





import pandas as pd
import numpy as np
import os
from datetime import datetime
import statsmodels.api as sm
from statsmodels.regression.rolling import RollingOLS
from multiprocessing import Pool

if __name__ == "__main__":
    from AssayingAnomalies import Config


def makeBetas(params):
    # Timekeeping
    print(f"\nNow working on making the betas. Run started at {datetime.now()}\n")

    crsp_path = params.crspFolder + os.sep
    daily_crsp_path = params.daily_crsp_folder + os.sep
    ff_path = params.ff_data_folder + os.sep

    "Start with the Frazzini - Pedersen betas"
    print(f"\nMaking Frazzini-Pedersen (2014) betas first.\n")

    # load necessary inputs
    dret = pd.read_parquet(daily_crsp_path + 'dret.parquet').astype(float)
    ff = pd.read_csv(crsp_path + 'ff.csv', index_col=0).astype(float)
    dff = pd.read_csv(ff_path + 'dff.csv', index_col=0).astype(float)
    ret = pd.read_csv(crsp_path + 'ret.csv', index_col=0).astype(float)

    # Convert to datetime indices
    dret.index = pd.to_datetime(dret.index, format='%Y-%m-%d')
    dff.index = pd.to_datetime(dff['dates'], format='%Y-%m-%d')
    ret.index = pd.to_datetime(ret.index.astype(int), format='%Y%m')

    # Store the excess stock returns
    rptdRf = np.tile(dff['rf'], (ret.shape[1], 1)).T
    dxret = dret - rptdRf

    # Store the 3-day excess returns and market returns
    # dret3 = (1 + dxret).rolling(window=3).cumprod() - 1
    dret3 = (1 + dxret).shift(1) * (1 + dxret).shift(2) * (1 + dxret).shift(3) - 1
    dmkt3 = (1 + dff['mkt']).shift(1) * (1 + dff['mkt']).shift(2) * (1 + dff['mkt']).shift(3) - 1

    # Compute rolling 1-year standard deviations
    stock_std = np.log(1 + dxret).rolling(window=252, min_periods=120).std()
    mkt_std = np.log(1 + dff['mkt']).rolling(window=252, min_periods=120).std()

    # Compute rolling 5-year correlations of 3-day excess returns with 3-day market returns
    corr = np.log(1 + dret3).rolling(window=252 * 5, min_periods=750).corr(np.log(1 + dmkt3))

    # Calculate beta = rho * (sigma_i / sigma_m)
    betas = corr * stock_std / np.tile(mkt_std, (dret.shape[1], 1)).T

    # Apply the shrinkage towards 1; Reduce the impact of extreme values.
    bfp = 0.6 * betas + 0.4

    bfp.columns = bfp.columns.astype(str)
    bfp.to_parquet(crsp_path + 'bfp.parquet')

    print("\nNow making Ivo's OLS, Dimson Correction, Vasicek shrinkage, 'standard', and Ivo Welch betas.\n")

    """
    Ivo's OLS benchmark - 1 year of daily data, 1 mkt lag, no shrinkage.
    Dimson correction - add one more lag of mkt.
    Vasicek shrinkage - similar to what LSY use.
    Standard - 1 market lag, shrinkage to 1.
    Ivo Welch's betas
    """
    # Calculate lagged market returns
    lagged_dmkt = pd.DataFrame(dff['mkt'], index=dff.index).shift(1)

    # Calculate the Ivo Welch winsorized excess returns
    lower_bound = np.multiply(-2, dff['mkt'])
    upper_bound = np.multiply(4, dff['mkt'])
    lower_bound.index = dxret.index
    upper_bound.index = dxret.index

    # Winsorize the excess returns data
    # Note: `clip` method is used for winsorization
    dxret_winsorized = dxret.clip(lower=lower_bound, upper=upper_bound, axis=0)

    # Create empty arrays to store the betas and ols errors
    betas_ols = pd.DataFrame(index=dxret.index, columns=dxret.columns)
    betas_ols_errors = pd.DataFrame(index=dxret.index, columns=dxret.columns)
    betas_dim = pd.DataFrame(index=dxret.index, columns=dxret.columns)
    betas_sw = pd.DataFrame(index=dxret.index, columns=dxret.columns)
    betas_vck = pd.DataFrame(index=dxret.index, columns=dxret.columns)

    # Create lagged excess return matrix
    # x1 = pd.DataFrame(sm.add_constant(dff['mkt']), index=dret.index)
    x1 = pd.DataFrame(sm.add_constant(dff['mkt']))
    x1.index = dret.index
    x2 = x1.merge(lagged_dmkt, how='outer', on=x1.index).iloc[:, 1:]
    x2.index = dret.index

    # Calculate OLS, Dimson , and Ivo Welch betas
    if not params.remote_or_not:
        for i, permno in enumerate(dret.columns):
            print(i, permno)
            if dxret.iloc[:, i].count() >= 120:
                # Benchmark OLS betas
                # betas_ols[:, i] = RollingOLS(dxret.iloc[:, i], x1, window=252, min_nobs=120).fit(params_only=True).params.iloc[:, 1]
                first_numeric = dxret.iloc[:, i].first_valid_index()
                last_numeric = dxret.iloc[:, i].last_valid_index()
                y1 = dxret.iloc[:, i][first_numeric:last_numeric]
                res = RollingOLS(y1, x1[first_numeric:last_numeric], window=min(y1.shape[0], 252), min_nobs=120).fit()
                betas_ols.loc[first_numeric:last_numeric, permno] = res.params.iloc[:, 1]
                betas_ols_errors.loc[:, i] = res.bse.iloc[:, 1]

                # Dimson correction - add 1 lag of mkt
                res = RollingOLS(y1, x2[first_numeric:last_numeric], window=min(y1.shape[0], 252), min_nobs=120).fit(params_only=True).params
                betas_dim.loc[first_numeric:last_numeric, i] = res.iloc[:, 1] + res.iloc[:, 2]

                # Ivo's betas - use the winsorized returns (dxretw)
                y2 = dxret_winsorized.iloc[:, i][first_numeric:last_numeric]
                betas_sw.loc[first_numeric:last_numeric, i] = RollingOLS(y2, x1[first_numeric:last_numeric], window=min(y1.shape[0], 252), min_nobs=120).fit(params_only=True).params.iloc[:, 1]
            else:
                continue

    else:
        def calculate_beta_for_column(permno, dxret_slice, dxret_winsorized_slice, x1, x2):
            print(f"Processing permno {permno}")
            # i = dxret_slice.name  # Assuming you pass the Series with its original name
            if dxret_slice.count() >= 120:
                first_numeric = dxret_slice.first_valid_index()
                last_numeric = dxret_slice.last_valid_index()
                y1 = dxret_slice[first_numeric:last_numeric]

                # Using min() because ocasionally the window is smaller than 252 but has 120 observations.
                res = RollingOLS(y1, x1.loc[first_numeric:last_numeric], window=min(y1.shape[0], 252), min_nobs=120).fit()
                ols_beta = res.params.iloc[:, 1]
                ols_error = res.bse.iloc[:, 1]

                res = RollingOLS(y1, x2.loc[first_numeric:last_numeric], window=min(y1.shape[0], 252), min_nobs=120).fit(
                    params_only=True).params
                dim_beta = res.iloc[:, 1] + res.iloc[:, 2]

                y2 = dxret_winsorized_slice[first_numeric:last_numeric]
                sw_beta = RollingOLS(y2, x1.loc[first_numeric:last_numeric], window=min(y1.shape[0], 252),
                                     min_nobs=120).fit(params_only=True).params.iloc[:, 1]

                return (permno, first_numeric, last_numeric, ols_beta, ols_error, dim_beta, sw_beta)
            else:
                return (permno, None, None, np.nan, np.nan, np.nan, np.nan)

        args = []
        for permno in dret.columns:
            dxret_slice = dxret[permno]
            dxret_winsorized_slice = dxret_winsorized[permno]
            args.append((permno, dxret_slice, dxret_winsorized_slice, x1, x2))

        with Pool(os.cpu_count()) as pool:
            results = pool.starmap(calculate_beta_for_column, args)

        # Process the results to update the DataFrames

        for i in range(len(results)):
            print(i)
            permno, first_numeric, last_numeric, ols_beta, ols_error, dim_beta, sw_beta = results[i]
            if first_numeric is not None:
                betas_ols.loc[first_numeric:last_numeric, permno] = ols_beta
                betas_ols_errors.loc[first_numeric:last_numeric, permno] = ols_error
                betas_dim.loc[first_numeric:last_numeric, permno] = dim_beta
                betas_sw.loc[first_numeric:last_numeric, permno] = sw_beta


    betas_ols = betas_ols.astype(float)
    betas_ols_errors = betas_ols_errors.astype(float)

    # Calculate the Vasicek shrinkage betas
    for i in range(betas_ols.shape[0]):  # Loop over time periods
        errors = betas_ols_errors.iloc[i, :].values
        betas = betas_ols.iloc[i, :].values
        # print(np.count_nonzero(~np.isnan(betas)))
        if np.count_nonzero(~np.isnan(betas)) > 2:
            sigmaSqI = errors ** 2
            sigmaSqT = np.nanvar(betas)
            wvck = sigmaSqT / (sigmaSqI + sigmaSqT)
            mean_beta = np.nanmean(betas)
            betas_vck.iloc[i, :] = wvck * betas + (1 - wvck) * mean_beta

    # Calculate standard betas
    betas_std = 0.6*betas_dim + 0.4

    # pd.DataFrame(betas_std, index=dret.index, columns=dret.columns).to_csv(crsp_path + 'betas_std.csv')
    # pd.DataFrame(betas_ols, index=dret.index, columns=dret.columns).to_csv(crsp_path + 'betas_ols.csv')
    # pd.DataFrame(betas_sw, index=dret.index, columns=dret.columns).to_csv(crsp_path + 'betas_sw.csv')
    # pd.DataFrame(betas_vck, index=dret.index, columns=dret.columns).to_csv(crsp_path + 'betas_vck.csv')
    # pd.DataFrame(betas_dim, index=dret.index, columns=dret.columns).to_csv(crsp_path + 'betas_dim.csv')

    betas_std.to_parquet(crsp_path + 'betas_std.parquet')
    betas_ols.to_parquet(crsp_path + 'betas_ols.parquet')
    betas_sw.to_parquet(crsp_path + 'betas_sw.parquet')
    betas_vck.to_parquet(crsp_path + 'betas_vck.parquet')
    betas_dim.to_parquet(crsp_path + 'betas_dim.parquet')

    print(f"Set-up is complete. Run ended at {datetime.now()}")

if __name__ == "__main__":
    params = Config.load_params()
    makeBetas(params)



from datetime import datetime
import requests
from io import BytesIO
import os
import numpy as np
import pandas as pd
from AssayingAnomalies.wrds_utilities import *

def remote_get_compustat_data(params, **kwargs):
    "Timekeeping"
    print(f"\nNow working on collecting data from COMPUSTAT. Run started at {datetime.now()}.\n")

    # Default parameter values
    p = {
        "delete_files_ok": True
    }

    # Update parameters with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Define the start and end parameters
    start = params.sample_start
    end = params.sample_end

    # Set the path to store compustat data.
    compFolder = params.compFolder + os.sep

    " First download the csv of necessary variable names"
    # location of csv file on velikov's github account
    url = 'https://raw.githubusercontent.com/velikov-mihail/AssayingAnomalies/main/Library%20Update/Inputs/COMPUSTAT%20Variable%20Names.csv'

    # tricks the host site into thinking the HTTP call is from user instead of machine
    headers = {"User-Agent": "Mozilla/5.0"}
    response = requests.get(url, headers=headers)
    # store the bytes object as dataframe
    COMPUSTAT_Variable_names = pd.read_csv(BytesIO(response.content))
    # save the dataframe
    COMPUSTAT_Variable_names.to_csv(compFolder + 'COMPUSTAT_Variable_names.csv')

    # create dataframes of annual and quarterly variable names
    COMPUSTAT_Variable_names_annual = np.array(COMPUSTAT_Variable_names.iloc[:, 0])
    # change the column names to lower case
    COMPUSTAT_Variable_names_annual = [i.lower() for i in COMPUSTAT_Variable_names_annual]
    # Adding additional columns that are needed in mergeCRSCOMP :TODO Note: Why is gvkey and datadate not included in the .csv file for annual variable names? Why is neither gvkey nor rdq included for quarterly?
    COMPUSTAT_Variable_names_annual.append('gvkey')
    COMPUSTAT_Variable_names_annual.append('datadate')

    # the entries after index 33 are NaN
    COMPUSTAT_Variable_names_quarterly = np.array(COMPUSTAT_Variable_names.iloc[:34, 1])
    # change the column names to lower case
    COMPUSTAT_Variable_names_quarterly = [i.lower() for i in COMPUSTAT_Variable_names_quarterly]
    # Adding additional columns that are needed in mergeCRSCOMP
    COMPUSTAT_Variable_names_quarterly.append('gvkey')
    COMPUSTAT_Variable_names_quarterly.append('rdq')

    # 'div' and 'xssd' columns are not found in funda so I am removing them
    COMPUSTAT_Variable_names_annual.remove('div')
    COMPUSTAT_Variable_names_annual.remove('xssd')


    # Create commands for custom sql queries and saving files
    commands = []
    file_locations = []
    sql_annual = "SELECT " + ", ".join(COMPUSTAT_Variable_names_annual) + " FROM comp.funda " \
                                                                          f"WHERE datadate>='01-01-{start}' " \
                                                                          f"and datadate<='12-31-{end}' "
    file_name_annual = "comp_funda.csv"
    file_location_annual = f"/scratch/rochester/{file_name_annual}"
    file_locations.append(file_location_annual)

    sql_quarterly = "SELECT " + ", ".join(COMPUSTAT_Variable_names_quarterly) + " FROM comp.fundq " \
                                                                                f"WHERE datadate>='01-01-{start}' " \
                                                                                f"and datadate<='12-31-{end}' "
    file_name_quarterly = "comp_fundq.csv"
    file_location_quarterly = f"/scratch/rochester/{file_name_quarterly}"
    file_locations.append(file_location_quarterly)

    command1 = f"testing = db.raw_sql(\"{sql_annual}\")"
    command2 = f"testing.to_csv(\"/scratch/rochester/{file_name_annual}\")"
    command3 = f"testing = db.raw_sql(\"{sql_quarterly}\")"
    command4 = f"testing.to_csv(\"/scratch/rochester/{file_name_quarterly}\")"

    commands.extend([command1, command2, command3, command4])

    # Start a new child process, and connect to the SSH server
    child = ssh_login(params)

    # Start a qrsh session
    start_qrsh(params, child)

    # Start Python session
    start_python_wrds(child)

    # Execute python commands on wrds cloud
    execute_commands(child, commands)

    # Exit all sessions
    exit_sessions(child)

    # Transfer the files and return a list of files that did not transfer
    print("Using SFTP to transfer files.")
    save_folder = params.compFolder
    problem_files = transfer_files(params, file_locations, save_folder)

    # Delete the files
    if p['delete_files_ok']:
        delete_files(params, file_locations, problem_files)

    # Print statement indicating if all files were completely transferred or not.
    if problem_files:
        print("The following files did not transfer: ")
        for file in problem_files:
            print(file)
    else:
        print("All files were successfully transferred and deleted from WRDS storage.")

    "Timekeeping"
    print(f"\nFinished collecting data from COMPUSTAT. Run ended at {datetime.now()}.\n")

    return


# from AssayingAnomalies import Config
# import AssayingAnomalies.Functions as aa
#
#
# "Create an instance of class 'Config' "
# params = Config()
#
# "Prompt the user to enter their parameters"
# params.prompt_user()
#
# "Create folders to store the downloaded data and created variables"
# params.make_folders()
#
# remote_get_compustat_data(params)



import numpy as np
import pandas as pd
import os
import statsmodels.api as sm

# TODO the alphas and excess returns are off by a factor of 100; everything else seems correct.
def estFactorRegs(params, pret, dates, factorModel=4, **kwargs):
    """
    This function performs factor model regressions on input portfolio returns
    and returns the results in the form of a dictionary containing various
    statistics, such as excess returns, t-stats, Sharpe ratios, information
    ratios, and other relevant metrics.

    Parameters:
    ----------
    pret : array-like
        A matrix of portfolio returns.
    dates : array-like
        A vector of dates corresponding to the portfolio returns.
    factorModel : int, optional (default=4)
        Factor model code (1: CAPM, 3: FF3 factor, 4: FF4 factor,
        5: FF5 factor, 6: FF6 factor).
    **kwargs : keyword arguments, optional
        - 'addLongShort': int (default=1)
            Flag equal to 0 or 1 indicating whether to add a long/short portfolio.

        - 'inputIsExcessReturn': int (default=0)
            Flag equal to 1 or 0 indicating whether the input is excess return.

    Returns:
    -------
    res : dict
        A dictionary containing the following results:
            - 'xret': array-like
                A vector of portfolio excess returns.

            - 'txret': array-like
                A vector of t-stats on the portfolio excess returns.

            - 'alpha': array-like
                A vector of alphas.

            - 'talpha': array-like
                A vector of t-stats of the alphas.

            - 'sharpe': array-like
                A vector of Sharpe ratios.

            - 'info': array-like
                A vector of information ratios.

            - 'pret': array-like
                A matrix containing the time-series of portfolio RAW returns.

            - 'factorModel': int
                A scalar indicating the factor model used.

            - 'nFactors': int
                A scalar indicating the number of factors.

            - 'loadings': array-like
                A matrix with factor loadings.

            - 'r2': array-like
                A vector of adjusted R-squared's from factor model regressions.

            - 'resid': array-like
                A matrix of residuals from factor model regressions (reduced size).
    """

    # Set dataPaths
    crspFolder = params.crspFolder + os.sep
    crspFolderDaily = params.daily_crsp_folder + os.sep

    # Default parameter values for **kwargs
    p = {
        'addLongShort': 1,
        'inputIsExcessReturn': 0,
    }

    # Update parameters with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Check if the factor model code is correct
    if factorModel not in {1, 2, 3, 4, 5, 6}:
        raise ValueError("Factor model must be 1, 3, 4, 5, 6, or a user-defined matrix.")

    # Store the number of portfolios
    numDates, numPtfs = pret.shape

    # Make sure dates are integers
    dates = dates.astype(int)

    # Check if daily dates
    if len(dates[0].astype(str)) == 6: # in this case, we are dealing with monthly data
        ff = pd.read_csv(crspFolder + 'ff.csv', index_col='dates')
        ff = ff.loc[dates] # align the factor model data with the portfolio return dates

    elif len(dates[0].astype(str)) == 8: # in this case we are dealing with daily data
        dff = pd.read_csv(crspFolderDaily + 'dff.csv', index_col='dates') # TODO:F I have not created daily data yet
        dff = dff.loc[dates] # align the factor model data with the portfolio return dates
        rf = dff.drf
        ffdates = dff.dffdates
        mkt = dff.dmkt
        smb = dff.dsmb
        smb2 = dff.smb2
        hml = dff.dhml
        umd = dff.dumd
        cma = dff.dcma
        rmw = dff.drmw

    else:
        raise ValueError("Unknown dates format for factor regressions")

    # Calculate excess returns if needed
    if p['inputIsExcessReturn'] == 0:
        pret = pret - ff['rf'].values[:, np.newaxis]

    # Check if we need to add the long/short portfolio
    if p['addLongShort'] != 0:
        long_short_ptf = pret[:, -1] - pret[:, 0]
        pret = np.column_stack((pret, long_short_ptf))
        numPtfs = numPtfs + 1

    # Set up the factor models
    if isinstance(factorModel, int):
        if factorModel == 1:
            x = sm.add_constant(ff['mkt'])
            labels = ['mkt']
        elif factorModel == 2:
            x = sm.add_constant(ff['mkt', 'smb'])
            labels = ['mkt', 'smb']
        elif factorModel == 3:
            x = sm.add_constant(ff[['mkt', 'smb', 'hml']])
            labels = ['mkt', 'smb', 'hml']
        elif factorModel == 4:
            x = sm.add_constant(ff[['mkt', 'smb', 'hml', 'umd']])
            labels = ['mkt', 'smb', 'hml', 'umd']
        elif factorModel == 5:
            x = sm.add_constant(ff[['mkt', 'smb', 'hml', 'rmw', 'cma']])
            labels = ['mkt', 'smb', 'hml', 'rmw', 'cma']
        elif factorModel == 6:
            x = sm.add_constant(ff[['mkt', 'smb', 'hml', 'umd', 'rmw', 'cma']])
            labels = ['mkt', 'smb', 'hml', 'umd', 'rmw', 'cma']
        else:
            raise ValueError("Invalid factor model. Must be 1, 3, 4, 5, or 6.")

    else: # user-defined factor model
        x = sm.add_constant(factorModel)

    # Run factor model regressions
    nFactors = x.shape[1] - 1
    alpha = np.zeros((numPtfs,))
    t_alpha = np.zeros((numPtfs,))
    t_factors = np.zeros((numPtfs, nFactors))
    r2 = np.zeros((numPtfs,))
    resid = np.zeros_like(pret)
    loadings = np.zeros((numPtfs, nFactors))

    for i in range(numPtfs):
        y = pret[:, i]
        # print(y.shape)
        model = sm.OLS(y, x, missing='drop').fit()
        # print("HERE")
        for j in range(nFactors):
            t_factors[i, j] = model.tvalues[j+1] # j+1 because the first t-stat is for the constant that we added above.
        alpha[i] = model.params[0]
        t_alpha[i] = model.tvalues[0]
        r2[i] = model.rsquared_adj
        resid[:len(model.resid), i] = model.resid # TODO:Note Because we have to drop the missing values we end up with a mismatch between the sizes of pret and resid. This is a temporary workaround but should be investigated further.
        loadings[i, :] = model.params[1:]

    # Calculate excess returns, t-stats, Sharpe ratios, and information ratios
    xret = np.nanmean(pret, axis=0)
    txret = xret / (np.nanstd(pret, axis=0) / np.sqrt(numDates))
    sharpe = xret / np.nanstd(pret, axis=0)
    info = alpha / np.nanstd(resid, axis=0)

    # Store the results in the output dictionary
    # TODO:N In MATLAB, factorLabels contains the factor labels, time series, loadings, and t-stats. I currently have
    #  them in separate fields.
    res = {
        'xret': xret,
        'txret': txret,
        'alpha': alpha,
        'talpha': t_alpha,
        'sharpe': sharpe,
        'info': info,
        'pret': pret,
        'factorModel': factorModel,
        'factorLabels': labels,
        'factors': np.array(x.iloc[:, 1:]),
        'nFactors': nFactors,
        'factorLoadings': loadings,
        'tfactors': t_factors,
        'r2': r2,
        'resid': resid
    }

    return res




import pandas as pd

def makeFF49Indus(SIC):
    def FF49_indus_mapping(SIC):
        # Agriculture
        if 100 <= SIC <= 299 or 700 <= SIC <= 799 or 910 <= SIC <= 919 or SIC == 2048:
            return 1

        # Food Products
        elif (2000 <= SIC <= 2046 or 2050 <= SIC <= 2063 or
              2070 <= SIC <= 2079 or 2090 <= SIC <= 2092 or
              SIC == 2095 or 2098 <= SIC <= 2099):
            return 2

        # Soda - Candy & Soda
        if ((2064 <= SIC <= 2068) or
                (SIC == 2086) or
                (SIC == 2087) or
                (SIC == 2096) or
                (SIC == 2097)):
            return 3

        # Beer - Beer & Liquor
        elif ((SIC == 2080) or
              (SIC == 2082) or
              (SIC == 2083) or
              (SIC == 2084) or
              (SIC == 2085)):
            return 4

        # Smoke - Tobacco Products
        elif 2100 <= SIC <= 2199:
            return 5

        # Toys - Recreation
        elif ((920 <= SIC <= 999) or
              (3650 <= SIC <= 3652) or
              (SIC == 3732) or
              (3930 <= SIC <= 3931) or
              (3940 <= SIC <= 3949)):
            return 6

        # Fun - Entertainment
        elif ((7800 <= SIC <= 7829) or
              (7830 <= SIC <= 7833) or
              (7840 <= SIC <= 7841) or
              (SIC == 7900) or
              (7910 <= SIC <= 7911) or
              (7920 <= SIC <= 7929) or
              (7930 <= SIC <= 7933) or
              (7940 <= SIC <= 7949) or
              (SIC == 7980) or
              (7990 <= SIC <= 7999)):
            return 7

        # Books - Printing and Publishing
        elif ((2700 <= SIC <= 2799) or
              (2770 <= SIC <= 2771) or
              (2780 <= SIC <= 2789)):
            return 8

        # Hshld - Consumer Goods
        elif ((SIC == 2047) or
              (2391 <= SIC <= 2392) or
              (2510 <= SIC <= 2599) or
              (2840 <= SIC <= 2844) or
              (3160 <= SIC <= 3172) or
              (3190 <= SIC <= 3199) or
              (SIC == 3229) or
              (SIC == 3260) or
              (3262 <= SIC <= 3263) or
              (SIC == 3269) or
              (3230 <= SIC <= 3231) or
              (3630 <= SIC <= 3639) or
              (3750 <= SIC <= 3751) or
              (SIC == 3800) or
              (3860 <= SIC <= 3873) or
              (3910 <= SIC <= 3915) or
              (3960 <= SIC <= 3962) or
              (SIC == 3991) or
              (SIC == 3995)):
            return 9

        # Clths - Apparel
        elif ((2300 <= SIC <= 2390) or
              (3020 <= SIC <= 3021) or
              (3100 <= SIC <= 3151) or
              (3963 <= SIC <= 3965)):
            return 10

        # Hlth - Healthcare
        elif 8000 <= SIC <= 8099:
            return 11

        # MedEq - Medical Equipment
        elif ((SIC == 3693) or
              (3840 <= SIC <= 3851)):
            return 12

        # Drugs - Pharmaceutical Products
        elif ((SIC >= 2830 and SIC <= 2831) or
              (2833 <= SIC <= 2836)):
            return 13

        # Chems - Chemicals
        elif ((2800 <= SIC <= 2829) or
              (2850 <= SIC <= 2879) or
              (2890 <= SIC <= 2899)):
            return 14

        # Rubbr - Rubber and Plastic Products
        elif ((SIC == 3031) or
              (SIC == 3041) or
              (3050 <= SIC <= 3099)):
            return 15

        # Txtls - Textiles
        elif ((2200 <= SIC <= 2299) or
              (2393 <= SIC <= 2399)):
            return 16

        # BldMt - Construction Materials
        elif ((800 <= SIC <= 899) or
              (2400 <= SIC <= 2499) or
              (2660 <= SIC <= 2661) or
              (2950 <= SIC <= 2952) or
              (3200 <= SIC <= 3299) or
              (3420 <= SIC <= 3452) or
              (3490 <= SIC <= 3499) or
              (SIC == 3996)):
            return 17

        # Cnstr - Construction
        elif ((1500 <= SIC <= 1549) or
              (1600 <= SIC <= 1799)):
            return 18

        # Steel - Steel Works Etc
        elif ((SIC == 3300) or
              (3310 <= SIC <= 3399)):
            return 19

        # FabPr - Fabricated Products
        elif ((SIC == 3400) or
              (3443 <= SIC <= 3479)):
            return 20

        # Mach - Machinery
        elif ((3510 <= SIC <= 3599) or
              (3580 <= SIC <= 3586) or
              (3589 <= SIC <= 3599)):
            return 21

        # ElcEq - Electrical Equipment
        elif ((SIC == 3600) or
              (3610 <= SIC <= 3699)):
            return 22

        # Autos - Automobiles and Trucks
        elif ((SIC == 2296) or
              (SIC == 2396) or
              (3010 <= SIC <= 3011) or
              (SIC == 3537) or
              (SIC == 3647) or
              (SIC == 3694) or
              (SIC == 3700) or
              (3710 <= SIC <= 3799)):
            return 23

        # Aero - Aircraft
        elif ((3720 <= SIC <= 3729)):
            return 24

        # Ships - Shipbuilding, Railroad Equipment
        elif ((3730 <= SIC <= 3731) or
              (3740 <= SIC <= 3743)):
            return 25

        # Guns - Defense
        elif ((3760 <= SIC <= 3769) or
              (SIC == 3795) or
              (3480 <= SIC <= 3489)):
            return 26

        # Gold - Precious Metals
        elif ((1040 <= SIC <= 1049)):
            return 27

        # Mines - Non-Metallic and Industrial Metal Mining
        elif ((1000 <= SIC <= 1039) or
              (1050 <= SIC <= 1119) or
              (1400 <= SIC <= 1499)):
            return 28

        # Coal - Coal
        elif ((1200 <= SIC <= 1299)):
            return 29

        # Oil - Petroleum and Natural Gas
        elif ((SIC == 1300) or
              (1310 <= SIC <= 1389) or
              (2900 <= SIC <= 2999)):
            return 30

        # Util - Utilities
        elif ((SIC == 4900) or
              (4910 <= SIC <= 4942)):
            return 31

        # Telcm - Communication
        elif ((SIC == 4800) or
              (4810 <= SIC <= 4899)):
            return 32

        # PerSv - Personal Services
        elif ((7020 <= SIC <= 7033) or
              (SIC == 7200) or
              (7210 <= SIC <= 7299) or
              (7395 <= SIC <= 7549) or
              (7600 <= SIC <= 7699) or
              (8100 <= SIC <= 8899) or
              (7510 <= SIC <= 7515)):
            return 33

        # BusSv - Business Services
        elif ((2750 <= SIC <= 2759) or
              (SIC == 3993) or
              (SIC == 7218) or
              (SIC == 7300) or
              (7310 <= SIC <= 8999) or
              (SIC == 7519) or
              (SIC == 8700) or
              (8710 <= SIC <= 8911) or
              (8920 <= SIC <= 8999) or
              (4220 <= SIC <= 4229)):
            return 34

        # Hardw - Computers
        elif ((3570 <= SIC <= 3579) or
              (3680 <= SIC <= 3695)):
            return 35

        # Softw - Computer Software
        elif ((7370 <= SIC <= 7373) or
              (SIC == 7375)):
            return 36

        # Chips - Electronic Equipment
        elif ((SIC == 3622) or
              (3661 <= SIC <= 3666) or
              (3669 <= SIC <= 3812)):
            return 37

        # LabEq - Measuring and Control Equipment
        elif ((SIC == 3811) or
              (3820 <= SIC <= 3827) or
              (3829 <= SIC <= 3839)):
            return 38
        # Paper - Business Supplies
        elif ((2520 <= SIC <= 2549) or
              (2600 <= SIC <= 2699) or
              (2760 <= SIC <= 2761) or
              (3950 <= SIC <= 3955)):
            return 39

        # Boxes - Shipping Containers
        elif ((2440 <= SIC <= 2449) or
              (2640 <= SIC <= 2659) or
              (3220 <= SIC <= 3221) or
              (3410 <= SIC <= 3412)):
            return 40

        # Trans - Transportation
        elif ((4000 <= SIC <= 4013) or
              (4040 <= SIC <= 4789)):
            return 41

        # Whlsl - Wholesale
        elif ((5000 <= SIC <= 5199)):
            return 42

        # Rtail - Retail
        elif ((5200 <= SIC <= 5999)):
            return 43

        # Meals - Restaurants, Hotels, Motels
        elif ((5800 <= SIC <= 7213)):
            return 44

        # Banks - Banking
        elif ((6000 <= SIC <= 6199)):
            return 45

        # Insur - Insurance
        elif ((6300 <= SIC <= 6411)):
            return 46

        # RlEst - Real Estate
        elif ((6500 <= SIC <= 6611)):
            return 47

        # Fin - Trading
        elif ((6200 <= SIC <= 6799)):
            return 48

        # Other - Almost Nothing
        elif ((4950 <= SIC <= 4991)):
            return 49

        else:
            return 0  # Return 0 or any other default value for unclassified industries

    FF49 = SIC.applymap(FF49_indus_mapping)

    data = {
        'shortName': ['Agric', 'Food', 'Soda', 'Beer', 'Smoke', 'Toys', 'Fun',
                      'Books', 'Hshld', 'Clths', 'Hlth', 'MedEq', 'Drugs', 'Chems',
                      'Rubbr', 'Txtls', 'BldMt', 'Cnstr', 'Steel', 'FabPr', 'Mach',
                      'ElcEq', 'Autos', 'Aero', 'Ships', 'Guns', 'Gold', 'Mines',
                      'Coal', 'Oil', 'Util', 'Telcm', 'PerSv', 'BusSv', 'Hardw',
                      'Softw', 'Chips', 'LabEq', 'Paper', 'Boxes', 'Trans', 'Whlsl',
                      'Rtail', 'Meals', 'Banks', 'Insur', 'RlEst', 'Fin', 'Other'],
        'longName': ['Agriculture', 'Food Products', 'Candy & Soda', 'Beer & Liquor',
                     'Tobacco Products', 'Recreation', 'Entertainment',
                     'Printing and Publishing', 'Consumer Goods', 'Apparel',
                     'Healthcare', 'Medical Equipment', 'Pharmaceutical Products',
                     'Chemicals', 'Rubber and Plastic Products', 'Textiles',
                     'Construction Materials', 'Construction', 'Steel Works Etc',
                     'Fabricated Products', 'Machinery', 'Electrical Equipment',
                     'Automobiles and Trucks', 'Aircraft', 'Shipbuilding, Railroad Equipment',
                     'Defense', 'Precious Metals', 'Non-Metallic and Industrial Metal Mining',
                     'Coal', 'Petroleum and Natural Gas', 'Utilities',
                     'Communication', 'Personal Services', 'Business Services',
                     'Computers', 'Computer Software', 'Electronic Equipment',
                     'Measuring and Control Equipment', 'Business Supplies',
                     'Shipping Containers', 'Transportation', 'Wholesale',
                     'Retail', 'Restaurants, Hotels, Motels', 'Banking',
                     'Insurance', 'Real Estate', 'Trading', 'Almost Nothing']
    }

    # Create the DataFrame of Industry names
    FF49_names = pd.DataFrame(data, index=range(1, 50))
    return FF49, FF49_names

# test1, test2 = makeFF49Indus(SIC)




from .runUnivSort import runUnivSort
import pandas as pd
import numpy as np
import os


def makeIndustryReturns(params, FFind):
    """
    Calculate and return industry-level returns for individual stocks.

    This function takes configuration parameters and a DataFrame specifying
    industry classifications for stocks. It reads return, market equity (ME),
    and date data from specified paths in 'params'. The function then runs
    a universe sort (using 'runUnivSort') based on these inputs, and computes
    industry-level returns. These returns are then assigned to individual stocks
    within each industry.

    Parameters:
    params (Config object): A configuration object containing various parameters
                            and file paths needed for processing.
    FFind (DataFrame): A DataFrame specifying the industry classification for each stock.

    Returns:
    tuple: A tuple containing two elements:
           - iret (DataFrame): A DataFrame with industry-level returns.
           - ireta (DataFrame): A DataFrame with industry-level returns assigned to individual stocks.

    The index of both 'iret' and 'ireta' DataFrames is based on the dates from the 'ret' DataFrame.
    The columns of 'ireta' correspond to individual stocks.

    Note:
    - This function relies on 'runUnivSort' to compute industry-level returns.
    - Industry classifications in 'FFind' should start from 1 and not 0.
    """

    comp_path = params.compFolder + os.sep
    crsp_path = params.crspFolder + os.sep

    ret = pd.read_csv(crsp_path + 'ret.csv', index_col=0).astype(float)
    me = pd.read_csv(crsp_path + 'me.csv', index_col=0).astype(float)
    dates = pd.read_csv(crsp_path + 'dates.csv', index_col=0).astype(float)

    res = runUnivSort(params, ret, FFind, me, dates, factorModel=1, addLongShort=0, printResults=1, plotFigure=0)

    iret = res['pret']
    nInds = iret.shape[1]
    nStocks = len(ret.columns)

    # Initialize ireta with NaNs
    ireta = pd.DataFrame(np.nan, index=ret.index, columns=ret.columns)

    # Assign industry returns to individual stocks
    for i in range(nInds):
        # Replicate industry returns across all stocks for the current industry
        rptdIret = pd.DataFrame(np.tile(iret[:, i], (len(ret.columns), 1)).T, index=ret.index, columns=ret.columns)

        # Apply the replicated returns to stocks belonging to the current industry
        industry_mask = (FFind == (i + 1))  # Adjusted for zero-based indexing in Python
        ireta[industry_mask] = rptdIret[industry_mask]

    return iret, ireta




import wrds
import numpy as np
import pandas as pd
import requests
import os
from io import BytesIO
from datetime import datetime
# import zipfile

def getCOMPUSTATData(params):
    # Timekeeping
    print(f"Began collecing COMPUSTAT data at {datetime.now()}")

    compFolder = params.compFolder + os.sep

    "Store the sample start and end dates"
    start = params.sample_start
    end = params.sample_end

    " First download the csv of necessary variable names"
    # location of csv file on velikov's github account
    url = 'https://raw.githubusercontent.com/velikov-mihail/AssayingAnomalies/main/Library%20Update/Inputs/COMPUSTAT%20Variable%20Names.csv'

    # tricks the host site into thinking the HTTP call is from user instead of machine
    headers = {"User-Agent": "Mozilla/5.0"}
    response = requests.get(url, headers=headers)
    # store the bytes object as dataframe
    COMPUSTAT_Variable_names = pd.read_csv(BytesIO(response.content))
    # save the dataframe
    COMPUSTAT_Variable_names.to_csv(compFolder + 'COMPUSTAT_Variable_names.csv')

    # create dataframes of annual and quarterly variable names
    COMPUSTAT_Variable_names_annual = np.array(COMPUSTAT_Variable_names.iloc[:, 0])
    # change the column names to lower case
    COMPUSTAT_Variable_names_annual = [i.lower() for i in COMPUSTAT_Variable_names_annual]
    # Adding additional columns that are needed in mergeCRSCOMP :TODO Note: Why is gvkey and datadate not included in the .csv file for annual variable names? Why is neither gvkey nor rdq included for quarterly?
    COMPUSTAT_Variable_names_annual.append('gvkey')
    COMPUSTAT_Variable_names_annual.append('datadate')

    # the entries after index 33 are NaN
    COMPUSTAT_Variable_names_quarterly = np.array(COMPUSTAT_Variable_names.iloc[:34, 1])
    # change the column names to lower case
    COMPUSTAT_Variable_names_quarterly = [i.lower() for i in COMPUSTAT_Variable_names_quarterly]
    # Adding additional columns that are needed in mergeCRSCOMP
    COMPUSTAT_Variable_names_quarterly.append('gvkey')
    COMPUSTAT_Variable_names_quarterly.append('rdq')

    "Establish connection to WRDS"
    db = wrds.Connection(wrds_username=params.username)

    # TODO:NOTE 'div' and 'xxsd' are not found in comp.funda so I am temporarily removing them"
    # for i in range(len(COMPUSTAT_Variable_names_annual)):
    #     print(i)
    #     db.get_table(library='comp', table='funda', columns=[COMPUSTAT_Variable_names_annual[i]], obs=10)
    COMPUSTAT_Variable_names_annual.remove('div')
    COMPUSTAT_Variable_names_annual.remove('xssd')

    "Create sql query to pass to db.raw_sql()"
    annual_sql = "SELECT " + ", ".join(COMPUSTAT_Variable_names_annual) + " FROM comp.funda " \
                                                                   f"WHERE datadate>='01-01-{start}' " \
                                                                   f"and datadate<='12-31-{end}' "

    quarterly_sql = "SELECT " + ", ".join(COMPUSTAT_Variable_names_quarterly) + " FROM comp.fundq " \
                                                                                f"WHERE datadate>='01-01-{start}' " \
                                                                                f"and datadate<='12-31-{end}' "

    "Get compustat annual data"
    # COMPUSTATQuery_annual = db.get_table(library='comp', table='funda', columns=COMPUSTAT_Variable_names_annual)
    COMPUSTATQuery_annual = db.raw_sql(annual_sql)
    # Save the file
    COMPUSTATQuery_annual.to_csv(compFolder + 'comp_funda.csv')

    "Get compustat quarterly data"
    # loop below tests if each column exists
    # for i in range(len(COMPUSTAT_Variable_names_quarterly)):
    #     print(i)
    #     db.get_table(library='comp', table='fundq', columns=[COMPUSTAT_Variable_names_quarterly[i]], obs=10)
    # COMPUSTAT_Variable_names_quarterly.remove('adjex')
    # COMPUSTATQuery_quarterly = db.get_table(library='comp', table='fundq', columns=COMPUSTAT_Variable_names_quarterly)
    COMPUSTATQuery_quarterly = db.raw_sql(quarterly_sql)
    COMPUSTATQuery_quarterly.to_csv(compFolder + 'comp_fundq.csv')

    db.close()

    print(f"Finished collecing COMPUSTAT data at {datetime.now()}")




import os
import sys
import pexpect
import paramiko


def ssh_login(params):
    host = 'wrds-cloud.wharton.upenn.edu'

    ssh_command = f"ssh {params.username}@{host}"
    child = pexpect.spawn(ssh_command, encoding='utf-8')
    child.expect('Password:', timeout=30)
    child.sendline(params.password)
    child.expect('\[.*@.*\]\$')
    child.sendline('1')
    print("Connected to WRDS.")
    return child


def start_qrsh(params, child):
    child.expect('\[.*@.*\]\$', timeout=30)
    child.sendline('qrsh')
    child.expect('password:', timeout=30)
    child.sendline(params.password)
    print("Started interactive wrds-cloud session")
    return child


def start_python_wrds(child, venv=False):
    child.expect('\[.*@.*\]\$', timeout=30)
    if venv:
        child.sendline('source virtualenvs/daily_environment/bin/activate')
        child.expect('\[.*@.*\]\$', timeout=30)
    child.sendline('python3')
    child.expect('>>>', timeout=30)
    child.sendline('import wrds')
    child.expect('>>>', timeout=30)
    child.sendline('db = wrds.Connection()')
    print("Started interactive python session.")
    return child


def execute_commands(child, commands, timeout=2400):
    child.logfile_read = sys.stdout
    for command in commands:
        child.expect('>>>', timeout=timeout)
        child.sendline(command)
    return child


def exit_sessions(child):
    child.expect('>>>', timeout=300)
    child.sendline('exit()')
    child.expect('\[.*@.*\]\$', timeout=30)
    child.sendline('exit')
    child.expect('\[.*@.*\]\$', timeout=30)
    child.sendline('exit')
    child.close()
    print("Exited all sessions.")


def transfer_files(params, file_locations, save_folder):
    host = 'wrds-cloud.wharton.upenn.edu'
    problem_files = []

    # Create a new SSH client
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    try:
        # Connect to the server
        ssh.connect(host, username=params.username, password=params.password)

        # Open an SFTP session
        sftp = ssh.open_sftp()

        # Transfer each file
        for file_location in file_locations:
            try:
                destination = f"{save_folder}" + os.sep + f"{file_location.split('/')[-1]}"
                print(destination)
                sftp.get(file_location, destination)
            except Exception as e:
                print(f"Error occurred while transferring {file_location}: {e}")
                problem_files.append(file_location)

        # Close the SFTP session
        sftp.close()

    except Exception as e:
        print(f"Error occurred while connecting: {e}")

    finally:
        # Close the SSH connection
        ssh.close()

    if problem_files:
        print("The following files did not transfer: ")
        for file in problem_files:
            print(file)

    return problem_files


def delete_files(params, file_locations, problem_files):
    host = 'wrds-cloud.wharton.upenn.edu'

    # Exclude problem files
    files_to_delete = [file for file in file_locations if file not in problem_files]

    # Generate command to delete multiple files at once instead of spawning new ssh session each time.
    command = f'ssh {params.username}@{host} rm ' + ' '.join(files_to_delete)

    # Delete all the files
    try:
        child = pexpect.spawn(command, encoding='utf-8')
        child.expect('Password: ')
        child.sendline(params.password)
        child.expect(pexpect.EOF)
        print("File deletion completed for: ", ', '.join(files_to_delete))
        child.close()
    except pexpect.exceptions.EOF:
        print("Connection reset while trying to delete files")


def create_virtual_environment(params):
    print("Creating virtual python environment.")

    # Start a new child process, and connect to the SSH server
    child = ssh_login(params)

    # Create virtual environment
    child.expect('\[.*@.*\]\$', timeout=30)
    child.sendline('python -m venv --system-site-packages ~/virtualenvs/daily_environment')
    child.logfile_read = sys.stdout

    # Activate environment
    child.expect('\[.*@.*\]\$', timeout=30)
    child.sendline('source virtualenvs/daily_environment/bin/activate')
    child.logfile_read = sys.stdout

    # pip install packages
    child.expect('\[.*@.*\]\$', timeout=30)
    child.sendline('pip install parquet pyarrow')
    child.logfile_read = sys.stdout

    # Deactivate environment
    child.expect('\[.*@.*\]\$', timeout=30)
    child.sendline('deactivate')
    child.logfile_read = sys.stdout

    child.close()
    print("\n\nCompleted virtual environment setup.\n")





from AssayingAnomalies.wrds_utilities.utilities import ssh_login, start_qrsh, start_python_wrds, exit_sessions,\
    execute_commands, transfer_files, delete_files


def remote_get_dsi(params, **kwargs):

    # Default keyword arguments
    p = {
        "delete_files_ok": True
    }

    # Update keyword arguments with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Define the columns to pull from WRDS
    columns_to_pull = ['vwretd']

    # Start a new child process, and connect to the SSH server
    child = ssh_login(params)

    # Start a qrsh session
    start_qrsh(params, child)

    # Start Python session
    start_python_wrds(params, child)

    # Create commands for custom sql query and saving files
    commands = []
    file_locations = []
    dsi_file_name = "crsp_dsi.csv"
    file_location_dsi = f'/scratch/rochester/{dsi_file_name}'
    file_locations.append(file_location_dsi)

    dsi_sql_string = "SELECT " + ", ".join(columns_to_pull) + f" FROM crsp.dsi " \
            f"WHERE date>='01-01-{params.sample_start}' " \
            f"and date<='12-31-{params.sample_end}' "


    command1 = f"testing = db.raw_sql(\"{dsi_sql_string}\")"
    command2 = f"testing.to_csv(\"/scratch/rochester/{dsi_file_name}\")"
    commands.extend([command1, command2])

    # Execute python commands on wrds cloud
    execute_commands(child, commands)

    # Exit all sessions
    exit_sessions(child)

    # Transfer the files and return a list of files that did not transfer
    print("Using SFTP to transfer files.")
    save_folder = params.daily_crsp_folder
    problem_files = transfer_files(params, file_locations, save_folder)

    # Delete the files
    if p['delete_files_ok']:
        delete_files(params, file_locations, problem_files)

    # Print statement indicating if all files were completely transferred or not.
    if problem_files:
        print("The following files did not transfer: ")
        for file in problem_files:
            print(file)
    else:
        print("All files were successfully transferred and deleted from WRDS storage.")

    return

# # from AssayingAnomalies import Config
# import AssayingAnomalies.Functions as AA
#
#
# "Create an instance of class 'Config' "
# params = Config()
#
# "Prompt the user to enter their parameters"
# params.prompt_user()
#
# "Create folders to store the downloaded data and created variables"
# params.make_folders()
#
# remote_get_dsi(params)



import pandas as pd
import os
from .makeFF10Indus import makeFF10Indus
from .makeFF17Indus import makeFF17Indus
from .makeFF49Indus import makeFF49Indus

# saveFolder = r'/home/jlaws13/PycharmProjects/AssayingAnomalies_root/Data/CRSP/'
def makeIndustryClassifications(params):
    """
    Purpose: This function creates and stores various industry classifications. Those include 4-digit SIC code, FF10,
    FF17, and FF49.
    :return:
    """
    saveFolder = params.crspFolder + os.sep

    siccd = pd.read_csv(saveFolder + 'siccd.csv', index_col=0)
    siccd.to_csv(saveFolder + 'SIC.csv')
    # Make the industry classifications
    FF10, FF10Names = makeFF10Indus(siccd)
    FF17, FF17Names = makeFF17Indus(siccd)
    FF49, FF49Names = makeFF49Indus(siccd)
    # TODO:F I have not made FF49 yet. It is very tedious.

    # Store the industry classification matrices
    FF10.to_csv(saveFolder + 'FF10.csv')
    FF10Names.to_csv(saveFolder + 'FF10Names.csv')
    FF17.to_csv(saveFolder + 'FF17.csv')
    FF17Names.to_csv(saveFolder + 'FF17Names.csv')
    FF49.to_csv(saveFolder + 'FF49.csv')
    FF49Names.to_csv(saveFolder + 'FF49Names.csv')

    return




import pandas as pd
import numpy as np
import os
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor
from .rank_with_nan import rank_with_nan
from .fill_months import fill_months
from .makePastPerformance import makePastPerformance
from .assignToPtf import assignToPtf
from .makeUnivSortInd import makeUnivSortInd


def makeNovyMarxVelikovAnomalies(params):
    # Timekeeping
    print(
        f"\nNow working on making anomaly signals from Novy-Marx and Velikov (RFS, 2016). Run started at {datetime.now()}\n")

    # Labels
    labels = ['size', 'value', 'grossProfitability', 'valProf', 'accruals', 'assetGrowth', 'investment', 'piotroski',
              'issuance', 'roe', 'distress', 'valMomProf', 'valMom', 'iVol', 'momentum', 'peadSUE', 'peadCAR3',
              'industryMomentum', 'industryRelativeReversal', 'highFrequencyCombo', 'reversals', 'seasonality',
              'industryRelativeReversalLowVol']


    # Store the general and daily CRSP data path
    crsp_path = params.crspFolder + os.sep
    daily_path = params.daily_crsp_folder + os.sep
    comp_path = params.compFolder + os.sep

    # Load a few basic variables
    ret = pd.read_csv(crsp_path + 'ret.csv', index_col=0).astype(float)
    me = pd.read_csv(crsp_path + 'me.csv', index_col=0).astype(float)
    dates = pd.read_csv(crsp_path + 'dates.csv', index_col=0).astype(int).values.flatten()

    # Store a few constants
    nMonths = len(ret.index)
    nStocks = len(ret.columns)

    # Initialize anomalies array
    anoms = np.full((nMonths, nStocks, 23), np.nan)

    # Size anomaly
    juneIndicator = (dates % 100 == 6)
    size = -me * juneIndicator[:, np.newaxis]  # turn juneIndicator into a col vector to multiply each col of me with.
    size[size == 0] = np.nan
    anoms[:, :, 0] = size

    # Value Anomaly
    value = pd.read_csv(comp_path + 'BM.csv', index_col=0).astype(float)
    anoms[:, :, 1] = value

    # Profitability Anomaly
    GP = pd.read_csv(comp_path + 'GP.csv', index_col=0).astype(float)
    AT = pd.read_csv(comp_path + 'AT.csv', index_col=0).astype(float)
    SIC = pd.read_csv(crsp_path + 'siccd.csv', index_col=0)
    FinFirms = (SIC >= 6000) & (SIC <= 6999) # creates a boolean mask that is 1 when the SIC is in the 6000's.
    gp = GP / AT
    gp[FinFirms == 1] = np.nan
    anoms[:, :, 2] = gp

    # ValProf Anomaly
    value[value == 0] = np.nan
    # gp_rank = np.apply_along_axis(rankdata, 1, gp)
    gp_rank = rank_with_nan(gp)
    # bm_rank = np.apply_along_axis(rankdata, 1, value)
    bm_rank = rank_with_nan(value)
    valProf = gp_rank + bm_rank
    anoms[:, :, 3] = valProf

    # Accruals
    ACT = pd.read_csv(comp_path + 'ACT.csv', index_col=0).astype(float)
    CHE = pd.read_csv(comp_path + 'CHE.csv', index_col=0).astype(float)
    LCT = pd.read_csv(comp_path + 'LCT.csv', index_col=0).astype(float)
    DLC = pd.read_csv(comp_path + 'DLC.csv', index_col=0).astype(float)
    TXP = pd.read_csv(comp_path + 'TXP.csv', index_col=0).astype(float)
    DP = pd.read_csv(comp_path + 'DP.csv', index_col=0).astype(float)

    dCA = ACT - ACT.shift(12)
    dCash = CHE - CHE.shift(12)
    dCL = LCT - LCT.shift(12)
    dSTD = DLC - DLC.shift(12)
    dTP = TXP - TXP.shift(12)
    Accruals = (dCA - dCash) - (dCL - dSTD - dTP) - DP
    accruals = -2 * Accruals / (AT + AT.shift(12))
    anoms[:, :, 4] = accruals

    # Asset Growth
    assetGrowth = -AT / AT.shift(12)
    assetGrowth[FinFirms == 1] = np.nan
    anoms[:, :, 5] = assetGrowth

    # Investment
    PPEGT = pd.read_csv(comp_path + 'PPEGT.csv', index_col=0).astype(float)
    INVT = pd.read_csv(comp_path + 'INVT.csv', index_col=0).astype(float)
    investment = -(PPEGT - PPEGT.shift(12) + INVT - INVT.shift(12)) / AT.shift(12)
    investment[FinFirms == 1] = np.nan
    anoms[:, :, 6] = investment

    # Piotroski's F-score (JAR 2000)
    IB = pd.read_csv(comp_path + 'IB.csv', index_col=0).astype(float)
    DLTT = pd.read_csv(comp_path + 'DLTT.csv', index_col=0).astype(float)
    SCSTKC = pd.read_csv(comp_path + 'SCSTKC.csv', index_col=0).astype(float)
    PRSTKCC = pd.read_csv(comp_path + 'PRSTKCC.csv', index_col=0).astype(float)
    # shrout = pd.read_csv(crsp_path + 'shrout.csv', index_col=0).astype(float)
    COGS = pd.read_csv(comp_path + 'COGS.csv', index_col=0).astype(float)
    REVT = pd.read_csv(comp_path + 'REVT.csv', index_col=0).astype(float)
    CFO = pd.read_csv(comp_path + 'CFO.csv', index_col=0).astype(float)

    ROA = IB / AT.shift(12)
    DTA = DLTT / AT
    ATL = ACT / LCT
    EqIss = SCSTKC.fillna(0) - PRSTKCC.fillna(0)
    # dshrout = shrout / shrout.shift(12) - 1
    GM = 1 - COGS / REVT
    ATO = REVT / AT.shift(12)

    # Financial Performance Signals
    F_ROA = (IB > 0).astype(int)
    F_CFO = (CFO > 0).astype(int)
    F_DROA = (ROA - ROA.shift(12) > 0).astype(int)
    F_ACCR = (CFO - IB > 0).astype(int)
    F_DLEV = ((DTA - DTA.shift(12) < 0) | ((DLTT == 0) & (DLTT.shift(12) == 0))).astype(int)
    F_ATL = (ATL - ATL.shift(12) > 0).astype(int)
    F_EQ = (EqIss <= 0).astype(int)
    F_GM = (GM - GM.shift(12) > 0).astype(int)
    F_ATO = (ATO - ATO.shift(12) > 0).astype(int)

    piotroski = F_ROA + F_CFO + F_DROA + F_ACCR + F_DLEV + F_ATL + F_EQ + F_GM + F_ATO

    # Available signals
    available = (F_ROA.notna() & F_CFO.notna() & F_DROA.notna() & F_ACCR.notna() &
                 F_DLEV.notna() & F_ATL.notna() & F_EQ.notna() & F_GM.notna() & F_ATO.notna()).sum(axis=1)
    piotroski[available < 9] = np.nan

    # Adjusting with Book-to-Market ratio
    bm_ranks = value.rank(axis=1)
    piotroski += bm_ranks / 1000

    # Store it
    anoms[:, :, 7] = piotroski

    # Issuance anomaly
    dashrout = pd.read_csv(crsp_path + 'dashrout.csv', index_col=0).astype(float)
    screen = np.ones(ret.shape)
    issuance = -dashrout * screen
    anoms[:, :, 8] = issuance

    # Return on Book Equity (ROE) anomaly
    IBQ = pd.read_csv(comp_path + 'IBQ.csv', index_col=0).astype(float)
    BEQ = pd.read_csv(comp_path + 'BEQ.csv', index_col=0).astype(float)
    roe = IBQ / BEQ.shift(3)
    anoms[:, :, 9] = roe

    # Distress risk
    IVOL = pd.read_parquet(crsp_path + 'IVOL.parquet').astype(float)
    prc = pd.read_csv(crsp_path + 'prc.csv', index_col=0).astype(float)
    ff = pd.read_csv(crsp_path + 'ff.csv', index_col=0).astype(float)
    # ACTQ = pd.read_csv(comp_path + 'ACTQ.csv', index_col=0).astype(float)
    # ATQ = pd.read_csv(comp_path + 'ATQ.csv', index_col=0).astype(float)
    CHEQ = pd.read_csv(comp_path + 'CHEQ.csv', index_col=0).astype(float)
    # DLCQ = pd.read_csv(comp_path + 'DLCQ.csv', index_col=0).astype(float)
    # DLTTQ = pd.read_csv(comp_path + 'DLTTQ.csv', index_col=0).astype(float)
    # IBQ = pd.read_csv(comp_path + 'IBQ.csv', index_col=0).astype(float)
    # LCTQ = pd.read_csv(comp_path + 'LCTQ.csv', index_col=0).astype(float)
    LTQ = pd.read_csv(comp_path + 'LTQ.csv', index_col=0).astype(float)
    NIQ = pd.read_csv(comp_path + 'NIQ.csv', index_col=0).astype(float)
    # PIQ = pd.read_csv(comp_path + 'PIQ.csv', index_col=0).astype(float)

    mktmat = (ff['mkt'] + ff['rf']).values.reshape(-1, 1) @ np.ones((1, me.shape[1]))
    temp = np.nansum(me, axis=1)
    memat = np.tile(temp, (me.shape[1], 1)).T  # repeats the temp series len(me.columns) times.
    coef = np.array([-20.264, 1.416, -7.129, 1.411, -0.045, -2.132, 0.075, -0.058])
    c = 2 ** (-1 / 3)
    NIfac = (1 - c ** 3) / (1 - c ** 12)

    # Calculating NIMTAAVG
    me.columns = me.columns.astype(float).astype(int).astype(str)  # :TODO fix how crsp columns labels are stored
    niq = NIQ / (LTQ + me)
    NIMTAAVG = np.full_like(me, np.nan, dtype=float)
    for i in range(9, len(me)):
        v1 = niq.iloc[i, :] + (c ** 3) * niq.iloc[i - 3, :] + (c ** 6) * niq.iloc[i - 6, :] + (
                c ** 9) * niq.iloc[i - 9, :]
        v2 = NIfac * v1
        NIMTAAVG[i, :] = NIfac * (niq.iloc[i, :] + (c ** 3) * niq.iloc[i - 3, :] + (c ** 6) * niq.iloc[i - 6, :] + (
                c ** 9) * niq.iloc[i - 9, :])

    # Calculating TLMTA
    TLMTA = (LTQ / (LTQ + me)).to_numpy()

    # Calculating EXRET
    EXRET = (np.log(1 + ret) - np.log(1 + mktmat)).to_numpy()
    XRfac = (1 - c) / (1 - c ** 12)
    EXRETAVG = np.full_like(me, np.nan, dtype=float)
    for i in range(12, len(me)):
        temp = EXRET[i - 1, :]
        for j in range(2, 13):
            temp += (c ** j) * EXRET[i - j, :]
        EXRETAVG[i, :] = XRfac * temp

    # Calculating SIGMA, RSIZE, and CASHMTA
    SIGMA = (np.sqrt(252) * IVOL).to_numpy()
    RSIZE = np.log(me / memat).to_numpy()
    CASHMTA = (CHEQ / (LTQ + me)).to_numpy()

    # Sum of negative book equity values
    BE = pd.read_csv(comp_path + 'BE.csv', index_col=0).astype(float)
    # negative_be_sum = (BE < 0).sum().sum()

    # Adjust BE and handle negative or zero values
    adjBE = 0.9 * BE + 0.1 * me
    adjBE[adjBE <= 0] = 0.001

    # Calculate MB and carry forward values
    MB = me.shift(6) / adjBE
    finite_mb_indices = MB.apply(lambda x: x.count(), axis=1) > 0

    for i in finite_mb_indices[finite_mb_indices].index:
        for j in range(1, 12):
            if i + j in MB.index:
                MB.loc[i + j] = MB.loc[i]

    MB = MB.to_numpy()
    # Calculate PRICE
    PRICE = prc.abs().clip(upper=15)
    PRICE[PRICE < 1] = np.nan
    PRICE[prc.isna()] = np.nan

    # Create helper function to winsorize the data
    def winsorize(series, percentile):
        lower = np.nanpercentile(series, percentile)
        upper = np.nanpercentile(series, 100 - percentile)
        return np.clip(series, lower, upper)

    # Applying winsorization and calculating 'distress'
    DISTRESS = sum(coef[i] * winsorize(variable, 5) for i, variable in
                   enumerate([NIMTAAVG, TLMTA, EXRETAVG, SIGMA, RSIZE, CASHMTA, MB, PRICE]))
    distress = -DISTRESS

    # Store the anomaly
    anoms[:, :, 10] = distress

    # Make ValMom & valMomPRof anoms
    R = pd.read_csv(crsp_path + 'R.csv', index_col=0).astype(float)
    # r_rank = np.apply_along_axis(rankdata, 1, R)
    r_rank = rank_with_nan(R)
    valMom = r_rank + fill_months(bm_rank, annual_or_quarterly='annual')  # :TODO Why are we filling months here?
    valMomProf = r_rank + fill_months(bm_rank, annual_or_quarterly='annual') + \
                 fill_months(gp_rank, annual_or_quarterly='annual')

    # Store the anomalies
    anoms[:, :, 11] = valMom
    anoms[:, :, 12] = valMomProf

    # Idiosyncratic volatility
    anoms[:, :, 13] = -pd.read_parquet(crsp_path + 'IffVOL3.parquet')

    # Momentum
    momentum = makePastPerformance(ret, 12, 1)
    anoms[:, :, 14] = momentum

    # PEAD (SUE)
    anoms[:, :, 15] = pd.read_csv(comp_path + 'SUE2.csv', index_col=0).astype(float)

    # PEAD (CAR3)
    anoms[:, :, 16] = pd.read_parquet(crsp_path + 'CAR3.parquet').astype(float)

    # Industry Momentum
    # FF49 = pd.read_csv(crsp_path + 'FF49.csv', index_col=0).astype(float)
    iFF49ret = pd.read_csv(crsp_path + 'iFF49ret.csv', index_col=0).astype(float)
    iFF49reta = pd.read_csv(crsp_path + 'iFF49reta.csv', index_col=0).astype(float)
    industryMomentum = assignToPtf(iFF49reta, np.sort(iFF49ret))
    # this expression creates a matrix where each row is a copy of the maximum value from the corresponding row in
    # industryMomentum, and the number of columns in this new matrix is the same as the number of columns in ret.
    max_values = np.tile(industryMomentum.max(axis=1), (ret.shape[1], 1)).T
    industryMomentum[industryMomentum == 0] = np.nan
    industryMomentum = industryMomentum / max_values
    anoms[:, :, 17] = pd.DataFrame(industryMomentum)

    # Industry-Relative Reversals
    IRR = rank_with_nan(iFF49reta.T - ret.T).T
    max_values = np.tile(np.nanmax(IRR, axis=1), (ret.shape[1], 1)).T
    industryRelativeReversal = IRR / max_values
    anoms[:, :, 18] = pd.DataFrame(industryRelativeReversal)

    # High Frequency Combo
    anoms[:, :, 19] = pd.DataFrame(industryMomentum + industryRelativeReversal)

    # Reversals
    anoms[:, :, 20] = -ret

    # Seasonality (Heston-Sadka)
    seasonality = pd.DataFrame(0, index=ret.index, columns=ret.columns)
    for i in range(1, 6):
        seasonality += ret.shift(periods=12 * i - 1)

    # Replace zeros with NaN and store
    seasonality.replace(0, np.nan, inplace=True)
    anoms[:, :, 21] = pd.DataFrame(seasonality)

    # IRRLowVol
    NYSE = pd.read_csv(crsp_path + 'NYSE.csv', index_col=0).astype(float)
    IffVol3 = pd.read_parquet(crsp_path + 'IffVOL3.parquet').astype(float)
    ind = makeUnivSortInd(IffVol3, 2, NYSE)
    industryRelativeReversalLowVol = IRR.copy()
    industryRelativeReversalLowVol[ind == 2] = np.nan
    anoms[:, :, 22] = pd.DataFrame(industryRelativeReversalLowVol)

    if not params.remote_or_not:
        for i, label in enumerate(labels):
            pd.DataFrame(anoms[:, :, i]).to_csv(crsp_path + f'nvmv_{label}.csv')

    else:
        def save_anoms(args):
            df, label = args
            print(f"Saving {label}")
            df.to_csv(crsp_path + f'nvmv_{label}.csv')
            print(f"Save completed for {label}.")

        tasks = [(pd.DataFrame(anoms[:, :, i]), labels[i]) for i in range(anoms.shape[2])]

        with ThreadPoolExecutor(max_workers=os.cpu_count()) as executor:
            executor.map(save_anoms, tasks)

    # :TODO:NOTE Here I have departed from the Matlab implemenation. Instead of saving a large panel dataframe, I am
    # # saving each dataframe independently. To turn the 3D array into a panel structure and save it, you can uncomment
    # # the section below and run it but it takes a significant amount of RAM.
    # Reshape and stack data into a table
    # nAnoms = len(labels)
    # # nObs = nStocks * nMonths
    # outputData = []
    #
    # permno_repeated = np.tile(ret.columns.astype(float), nMonths)
    # dates_repeated = np.tile(ret.index.astype(int), nStocks)
    #
    # outputData.append(permno_repeated)
    # outputData.append(dates_repeated)
    #
    # # Append all the anomaly columns
    # for i in range(nAnoms):
    #     outputData.append(anoms[:, :, i].flatten())
    #
    # # Create DataFrame
    # col_names = ['permno', 'date'] + labels
    # outputData = pd.DataFrame(outputData, index=col_names).T
    #
    # # Drop rows that are entirely nan values
    # outputData.dropna(how='all', inplace=True, subset=labels)
    #
    # # Save
    # outputData.columns = outputData.columns.astype(str)
    # outputData.to_parquet(crsp_path + 'novyMarxVelikovAnomalies.parquet')

    # Timekeeping
    print(f"\nAnomaly signal run ended, data exported at {datetime.now()}\n")

    return


    # # Example Data
    # nMonths = 2  # For example, two months
    # nStocks = 3  # For example, three stocks
    # nAnoms = 3  # Three anomalies
    #
    # # Simulate stock identifiers and dates
    # permno = np.array([101, 102, 103])  # Stock identifiers
    # dates = pd.date_range(start='2020-01-01', periods=nMonths, freq='M')  # Dates for two months
    #
    # # Simulate anomaly data
    # anoms = np.random.rand(nMonths, nStocks, nAnoms)  # Random data for anomalies
    #
    # # Reshape and stack data into a table
    # nObs = nStocks * nMonths
    # outputData = []
    #
    # # Add permno and dates columns
    # permno_repeated = np.tile(permno, nMonths)
    # dates_repeated = np.repeat(dates, nStocks)
    # outputData.append(permno_repeated)
    # outputData.append(dates_repeated)
    #
    # # Append all the anomaly columns
    # for i in range(nAnoms):
    #     outputData.append(anoms[:, :, i].flatten())
    #
    # # Create DataFrame
    # outputData = pd.DataFrame(outputData).T
    # outputData.columns = ['permno', 'dates', 'anom1', 'anom2', 'anom3']
    #
    # # Remove rows with NaNs in all anomaly columns
    # outputData = outputData.dropna(subset=['anom1', 'anom2', 'anom3'], how='all')




import pandas as pd
import numpy as np
import os
from datetime import datetime
from AssayingAnomalies.config import Config

def makeCorwinSchultz():
    """
    This function estimates bid-ask spreads using the Corwin-Schultz estimator, a methodology that infers the bid-ask
    spread from daily high and low prices. It applies adjustments for negative prices and overnight price changes and
    computes variables such as lclose, pchg, lhi, llo, beta, gamma, alpha, and the final spread estimate 's'. It sets
    negative spread estimates to zero and computes monthly averages of the high-low spread estimate.

    Parameters:
    ----------
    params : object
        An object containing the following attributes:
        - crspFolder: str
            The file path to the CRSP data folder.
        - daily_crsp_folder: str
            The file path to the daily CRSP data folder.

    Returns:
    -------
    res : pd.DataFrame
        A DataFrame containing the monthly average of the high-low spread estimate for each stock.

    References:
    -----------
    1. Corwin, S. and P. Schultz, 2012, A simple way to estimate bid-ask spreads from daily high and low prices,
    Journal of Finance.
    2. Chen, A. and M. Velikov, 2021, Zeroing in on the expected return on anomalies, Journal of Financial and
    Quantitative Analysis (JFQA).
    3. Novy-Marx, R. and M. Velikov, 2023, Assaying Anomalies, working paper.
    """
    print(f"Making Corwin Schultz (2012) bid-ask spread estimates. Run started at {datetime.now()}")

    params = Config.load_params()
    crsp_path = params.crspFolder + os.sep
    daily_crsp_path = params.daily_crsp_folder + os.sep

    # Load the necessary variables
    # dbidlo = pd.read_csv(daily_crsp_path + 'dbidlo.csv', index_col=0).astype(float)
    # daskhi = pd.read_csv(daily_crsp_path + 'daskhi.csv', index_col=0).astype(float)
    ddates = pd.read_csv(daily_crsp_path + 'ddates.csv', index_col=0).astype(int)
    # dprc = pd.read_csv(daily_crsp_path + 'dprc.csv', index_col=0).astype(float)
    # dcfacpr = pd.read_csv(daily_crsp_path + 'dcfacpr.csv', index_col=0).astype(float)
    # ret = pd.read_csv(crsp_path + 'ret.csv', index_col=0).astype(float)
    dates = pd.read_csv(crsp_path + 'dates.csv', index_col=0).astype(int)
    dbidlo = pd.read_parquet(daily_crsp_path + 'dbidlo.parquet').astype(float)
    daskhi = pd.read_parquet(daily_crsp_path + 'daskhi.parquet').astype(float)
    dprc = pd.read_parquet(daily_crsp_path + 'dprc.parquet').astype(float)
    dcfacpr = pd.read_parquet(daily_crsp_path + 'dcfacpr.parquet').astype(float)
    permno = pd.read_csv(crsp_path + 'permno.csv', index_col=0).astype(float)

    # If the price was ever negative, replace the hi and low values with np.nan
    daskhi.iloc[dprc < 0] = np.nan
    dbidlo.iloc[dprc < 0] = np.nan

    # Take the absolute value of daily prices
    dprc = dprc.abs()

    # Compute lclose
    lclose = np.divide(np.multiply(dprc.shift(1).fillna(np.nan), dcfacpr), dcfacpr.shift(1).fillna(np.nan))
    # Page 726: "If the day t+1 low is above the day t close, we assume that the price rose overnight from the close to
    # the day t + 1 low and decrease both the high and low for day t + 1 by the amount of the overnight change when
    # calculating spreads. Similarly, if the day t+1 high is below the day t close, we assume the price fell overnight
    # from the close to the day t + 1 high and increase the day t + 1 high and low prices by the amount of this
    # overnight decrease."

    # Initialize pchg as a DataFrame of zeros with the same size as dprc
    pchg = pd.DataFrame(np.zeros_like(dprc.values), index=dprc.index, columns=dprc.columns)

    # Calculate overnight change
    condition_dbidlo = (dbidlo > lclose).values
    condition_daskhi = (daskhi < lclose).values
    pchg.values[condition_dbidlo] = (dbidlo - lclose).values[condition_dbidlo]
    pchg.values[condition_daskhi] = (daskhi - lclose).values[condition_daskhi]

    # Adjust hi and lo
    dbidlo -= pchg
    daskhi -= pchg

    # Compute lhi and llo
    lhi = np.divide(np.multiply(daskhi.shift(1).fillna(np.nan), dcfacpr), dcfacpr.shift(1).fillna(np.nan))
    llo = np.divide(np.multiply(dbidlo.shift(1).fillna(np.nan), dcfacpr), dcfacpr.shift(1).fillna(np.nan))

    #=================================================================================================================#
    #============================================== Acual spread estimation ==========================================#
    #=================================================================================================================#

    # Calculate beta
    beta = (np.log(daskhi / dbidlo)) ** 2 + (np.log(lhi / llo)) ** 2

    # NaN assignment for beta
    beta[(daskhi <= 0) | (dbidlo <= 0)] = np.nan

    # Calculate gamma
    gamma = (np.log(np.maximum(daskhi, lhi) / np.minimum(dbidlo, llo))) ** 2

    # NaN assignment for gamma
    gamma[np.minimum(dbidlo, llo) <= 0] = np.nan

    # Define the constant
    const = 3 - 2 * np.sqrt(2)

    # Calculate alpha according to Equation (18)
    alpha = (np.sqrt(2 * beta) - np.sqrt(beta)) / const - np.sqrt(gamma / const)

    # Calculate s according to Equation (14)
    s = 2 * (np.exp(alpha) - 1) / (1 + np.exp(alpha))
    #==================================================================================================================
    # Section II.C
    #==================================================================================================================
    # Page 727: "The high-low estimator assumes that the expectation of a stock's true variance over a 2-day period is
    # twice as large as the expectation of the variance over a single day. Even if this is true in expectation, the
    # observed 2-day variance may be more than twice as large as the single-day variance during volatile periods, in
    # cases with a large overnight price change, or when the total return over the 2-day period is large relative to the
    # intraday volatility. If the observed 2-day variance is large enough, the high-low spread estimate will be
    # negative. For most of the analysis to follow, we set all negative 2-day spreads to zero before calculating monthly
    # averages.

    # Set negative values in s to 0
    s[s < 0] = 0

    # Initialize hl with NaNs and determine the number of months
    # hl = pd.DataFrame(np.nan, index=dates.values.flatten(), columns=s.columns)
    # nMonths = len(dates)
    "The commented out section below closely follows the Matlab procedure but is very slow. The .resample() method is" \
    "MUCH faster."
    #
    # # Loop through the months
    # for i in range(nMonths):
    #     print(i)
    #     # Find indices for the current month
    #     ind = (ddates // 100 == dates.iloc[i])
    #     print('a')
    #     # Determine columns with at least 12 finite values
    #     hor_index = s[ind].count() >= 12
    #     print('b')
    #     # Calculate mean for these columns
    #     hl.iloc[i, hor_index] = s[ind].mean()
    #     print('c')
    # # Remove any imaginary numbers (if any)
    # hl = hl.apply(np.real)
    #
    # # Convert ddatest to a monthly format for grouping
    # ddates_monthly = ddates // 100
    #
    # # Perform operations for each month
    # for count, date in enumerate(hl.index):
    #     print(count)
    #     # Filter data for the current month
    #     monthly_data = s[ddates_monthly == date]
    #
    #     # Check if there are at least 12 finite values in each column
    #     valid_columns = monthly_data.count() >= 12
    #
    #     # Calculate mean for valid columns
    #     hl.loc[date, valid_columns] = monthly_data.mean()
    s.index = ddates.values.flatten()
    s.index = pd.to_datetime(s.index, format='%Y%m%d')
    hl = s.resample('M').mean()

    print(f"Finished making Corwin Schultz (2012) bid-ask spread estimates at {datetime.now()}")

    return hl



from glob import glob
from scipy.stats.mstats import winsorize
from multiprocessing import Pool
from .makeGibbs import makeGibbs
from .makeCorwinSchultz import makeCorwinSchultz
from .makeAbdiRanaldi import makeAbdiRanaldi
from .makeKyleObizhaeva import makeKyleObizhaeva
from .make_hf_effective_spreads import *
from .rank_with_nan import rank_with_nan
from .fill_missing_tcosts import fill_missing_tcosts


def parallel_make_trading_costs(function_name):
    if function_name == 'hl':
        return 'hl', makeCorwinSchultz()
    elif function_name == 'chl':
        return 'chl', makeAbdiRanaldi()
    elif function_name == 'vov':
        return 'vov', makeKyleObizhaeva()
    else:
        raise ValueError("Invalid function name")


def makeTradingCosts(params):
    # Timekeeping
    print(f"\nNow working on creating the transaction costs. Run started at {datetime.now()}\n")

    # Store the general and daily CRSP data path
    crsp_path = params.crspFolder + os.sep
    daily_path = params.daily_crsp_folder + os.sep
    hf_path = params.hf_effective_spreads_folder + os.sep

    # Store the tcost types
    # tcostsType = params.tcostsType
    tcostsType = 'full'

    # Check if correct tcosts input selected
    if tcostsType not in ['full', 'lf_combo', 'gibbs']:
        print(f"params.tcostsType is {tcostsType} but should be one of the folowing: \"full\", \"lf_combo\", \"gibbs\"")

    # Initialize dictionary to hold trading costs. CorwinSchultz = hl, AbdiRanaldi = chl, KyleObizhaeva = vov
    effSpreadStruct = {'gibbs': None,
                       'hl': None,
                       'chl': None,
                       'vov': None,
                       'hf_spreads_ave': None
                       }

    "Check for Gibbs file"
    # Construct the file search pattern
    search_pattern = os.path.join(params.data_folder, '**', 'crspgibbs.csv')

    # Find all files matching the pattern
    gibbs_file_list = glob(search_pattern, recursive=True)  # recursive=True searches the subdirectories as well

    # Check if any files were found
    if not gibbs_file_list:
        raise FileNotFoundError('Gibbs input file does not exist. Gibbs trading cost estimate cannot be constructed.')
    else:
        file_path = gibbs_file_list[0]

    "Create Gibbs spreads"
    # path to file with Hasbrouck effective spread estimates
    effSpreadStruct['gibbs'] = makeGibbs(params, file_path)

    if tcostsType in ['lf_combo', 'full']:
        if not params.remote_or_not:
            effSpreadStruct['hl'] = makeCorwinSchultz()
            effSpreadStruct['chl'] = makeAbdiRanaldi()
            effSpreadStruct['vov'] = makeKyleObizhaeva()

        else:
            functions_to_run = ['hl', 'chl', 'vov']

            # Create a pool of workers and run the functions in parallel
            with Pool(processes=min(3, os.cpu_count())) as pool:
                results = pool.map(parallel_make_trading_costs, functions_to_run)

            # Update effSpreadStruct with the results
            for key, value in results:
                effSpreadStruct[key] = value

    if tcostsType == 'full':
        search_pattern = os.path.join(params.data_folder, '**', 'hf_monthly_pre_2003.csv')
        hf_file_list = glob(search_pattern, recursive=True)
        if not hf_file_list:
            raise FileNotFoundError('High-frequency trading cost input file does not exist. High-frequency trading '
                                    'cost estimate cannot be constructed prior to 2003.')
        else:
            get_hf_spreads_data(params)
            make_hf_effective_spreads(params)
            effSpreadStruct['hf_spreads_ave'] = extend_hf_effective_spreads(params)


    # Winsorize by keeping anything between 0 and 99.9th percentiles.
    for key in effSpreadStruct.keys():
        flat_values = effSpreadStruct[f'{key}'].to_numpy().flatten()
        winsorized_values = winsorize(flat_values, limits=[0, 0.001], nan_policy='omit')
        effSpreadStruct[f'{key}'] = pd.DataFrame(winsorized_values.reshape(effSpreadStruct[f'{key}'].shape))

    # Check if we need to adjust some of the tcost measures
    if tcostsType.lower() == 'gibbs':
        # No need to worry about the rest
        tcosts_raw = effSpreadStruct['gibbs'] / 2
        # tcosts_raw.to_csv(crsp_path + 'tcosts_raw.csv')
        tcosts_raw.columns = tcosts_raw.columns.astype(str)
        tcosts_raw.to_parquet(crsp_path + 'tcosts_raw.parquet')
    else:
        exchcd = pd.read_csv(crsp_path + 'exchcd.csv', index_col=0).astype(float)
        exchcd.index = pd.to_datetime(exchcd.index.astype(int), format='%Y%m')

        # Excluding NASDAQ stocks prior to 1983 for Gibbs and VoV
        # Create a mask for dates before 1983 and exchange code being 3
        years = np.tile(exchcd.index.year, (len(exchcd.columns), 1)).T
        mask = (exchcd == 3) & (years < 1983)
        temp = effSpreadStruct['gibbs'].to_numpy()
        temp[mask] = np.nan
        effSpreadStruct['gibbs'] = temp
        temp = effSpreadStruct['vov'].to_numpy()
        temp[mask] = np.nan
        effSpreadStruct['vov'] = temp

        # Excluding Nasdaq stocks prior to 1993 for HL and CHL
        mask = (exchcd == 3) & (years < 1993)
        temp = effSpreadStruct['hl'].to_numpy()
        temp[mask] = np.nan
        effSpreadStruct['hl'] = temp
        temp = effSpreadStruct['chl'].to_numpy()
        temp[mask] = np.nan
        effSpreadStruct['chl'] = temp

        # Excluding AMEX stocks prior to 1962 for all
        mask = (exchcd == 3) & (years < 1962)
        for key in ['gibbs', 'hl', 'chl', 'vov']:
            temp = effSpreadStruct[key]
            temp[mask] = np.nan
            effSpreadStruct[key] = pd.DataFrame(temp)

        # In the matlab code, reshapedEffSpreadRaw is a long vector containing the average trading cost for each
        # permno for each month across the 4 lf measures. However, this is then reshaped back to a matrix. I
        EffSpreadRaw = pd.concat([effSpreadStruct['gibbs'], effSpreadStruct['hl'], effSpreadStruct['chl'],
                                  effSpreadStruct['vov']], axis=0).groupby(level=0).mean()

        # Update the average trading costs with the high frequency estimates where available.
        mask = np.isfinite(effSpreadStruct['hf_spreads_ave'])
        EffSpreadRaw[mask] = effSpreadStruct['hf_spreads_ave']

        # Need to divide the effective spreads by 2, because this is the tcost measure (half-spread!)
        tcosts_raw = EffSpreadRaw / 2

        # Store the raw tcosts
        # tcosts_raw.to_csv(crsp_path + 'tcosts_raw.csv')
        tcosts_raw.columns = tcosts_raw.columns.astype(str)
        tcosts_raw.to_parquet(crsp_path + 'tcosts_raw.parquet')

    # Fill in missing trading costs
    # Load variables we need
    me = pd.read_csv(crsp_path + 'me.csv', index_col=0).astype(float)
    rme = rank_with_nan(me)

    # Load IffVOL3 and rIVOL outside the function
    IffVOL3 = pd.read_parquet(crsp_path + 'IffVOL3.parquet')
    rIVOL = rank_with_nan(IffVOL3)
    if not params.remote_or_not:
        tcosts = fill_missing_tcosts(tcosts_raw, rIVOL, rme)

    else:
        # Split tcosts, rme, and rIVOL into chunks
        tcosts_chunks = np.array_split(tcosts_raw, os.cpu_count())
        rme_chunks = np.array_split(rme, os.cpu_count())
        rIVOL_chunks = np.array_split(rIVOL, os.cpu_count())
        chunks = zip(tcosts_chunks, rme_chunks, rIVOL_chunks)

        with Pool(os.cpu_count()) as pool:
            results = pool.starmap(fill_missing_tcosts, chunks)

        tcosts = pd.concat(results)

    # tcosts.to_csv(crsp_path + 'tcosts.csv')
    tcosts.columns = tcosts.columns.astype(str)
    tcosts.to_parquet(crsp_path + 'tcosts.parquet')

    # Do the FF trading costs calculation here too # :TODO Make the makeFFTcosts function.
    # makeFFTcosts()

    # Timekeeping
    print(f"Trading costs construction run ended at {datetime.now()}")


# # Uncomment to test
# import numpy as np
# import pandas as pd
# import os
# from glob import glob
# from scipy.stats.mstats import winsorize
# from datetime import datetime
# from multiprocessing import Pool
# from AssayingAnomalies.Functions.makeGibbs import makeGibbs
# from AssayingAnomalies.Functions.makeCorwinSchultz import makeCorwinSchultz
# from AssayingAnomalies.Functions.makeAbdiRanaldi import makeAbdiRanaldi
# from AssayingAnomalies.Functions.makeKyleObizhaeva import makeKyleObizhaeva
# from AssayingAnomalies.Functions.make_hf_effective_spreads import *
# from AssayingAnomalies.Functions.rank_with_nan import rank_with_nan
# from AssayingAnomalies.Functions.fill_missing_tcosts import fill_missing_tcosts


if __name__ == '__main__':
    params = src.AssayingAnomalies.config.Config.load_params()
    makeTradingCosts(params)
